# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" API connector """

import logging

from eessh.cache import xdgcache as region
from eessh.config import BooleanOption, PasswordOption, StringOption, config
from eessh.legacy import LegacyConnector

log = logging.getLogger(__name__)


class API:
    """API connector"""

    def __init__(self):
        self.__connector = None

    @property
    def ready(self):
        """Check if API connector is ready"""
        return self.__connector is not None

    def get_customers(self, *args, **kwargs):
        """Retrieve customers info"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.get_customers(*args, **kwargs)

    def get_data(self, *args, **kwargs):
        """Retrieve server extended info"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.get_data(*args, **kwargs)

    def get_ipmi(self, *args, **kwargs):
        """Retrieve server IPMI info"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.get_ipmi(*args, **kwargs)

    def get_server(self, *args, **kwargs):
        """Retrieve server info"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.get_server(*args, **kwargs)

    def get_servers(self, *args, **kwargs):
        """Retrieve customer servers"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.get_servers(*args, **kwargs)

    def get_tickets(self, *args, **kwargs):
        """Retrieve server tickets"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.get_tickets(*args, **kwargs)

    def update_data(self, *args, **kwargs):
        """Update server info"""
        assert self.__connector is not None, "API not initialized"

        with self.__connector as connector:
            return connector.update_data(*args, **kwargs)

    @region.cache_on_arguments_if_not_refresh(leave_refresh_kwarg=True)
    def get_customers_uid(self, refresh=False):
        """Retrieve list of customers UID"""
        return [customer.get("customer_uid") for customer in self.get_customers(refresh=refresh)]

    @region.cache_on_arguments_if_not_refresh(leave_refresh_kwarg=True)
    def get_customer_servernames(self, customer_uid, refresh=False):
        """Retrieve list of customer server hostnames"""
        return [
            server.get("hostname")
            for server in self.get_servers(customer_uid=customer_uid, refresh=refresh)
        ]

    @staticmethod
    def invalidate_cache():
        """Invalidate API cache"""
        region.invalidate()

    def initialize(self, loaded_config):
        """Initialize connector"""

        if loaded_config.get("api", "legacy") is True:
            class_ = LegacyConnector
        else:
            log.error("Only legacy mode is supported for now")
            return False

        url = loaded_config.get("api", "url")
        if url is None:
            log.error("Missing API url, please check configuration")
            return False

        options = {
            option: loaded_config.get("api", option) for option in ("debug", "password", "username")
        }

        self.__connector = class_(url, **options)

        return True


# Initialize API object
api = API()

# Register config section, options and config loaded callback
section = config.add_section(
    "api", comment="API connection", order=1, loaded_callback=api.initialize
)
section.add_option(StringOption, "url", default="https://eemyle.example.tld/", comment="API URL")
section.add_option(BooleanOption, "legacy", default=True, comment="Enable/disable API legacy mode")
section.add_option(StringOption, "username", default="foo", comment="API username")
section.add_option(
    PasswordOption,
    "password",
    default="keyring",
    comment='API password (set to "keyring" to use XDG keyring)',
    username_option="username",
    keyring_value="keyring",
)
section.add_option(BooleanOption, "debug", default=False, comment="Enable API debugging")
