# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Cache manager """

import logging
import os

from dogpile.cache import make_region
from dogpile.util import compat

from eessh.config import DEFAULT_CACHE_DIRPATH, IntegerOption, StringOption, config

log = logging.getLogger(__name__)
DEFAULT_CACHE_EXPIRATION = 3600 * 24 * 7


def initialize(loaded_config):
    """Configuration initialization callback"""
    dirpath = loaded_config.get("cache", "dirpath")
    if not os.path.isdir(dirpath):
        log.warning("Creating cache directory (%s)", dirpath)
        os.makedirs(dirpath)

    xdgcache.configure(
        "dogpile.cache.dbm",
        expiration_time=loaded_config.get("cache", "expiration"),
        arguments={
            "filename": os.path.join(dirpath, "default"),
        },
    )


# Register config section & options
section = config.add_section("cache", comment="Caching", order=99, loaded_callback=initialize)
section.add_option(
    StringOption, "dirpath", default=DEFAULT_CACHE_DIRPATH, comment="Cache directory path"
)
section.add_option(
    IntegerOption,
    "expiration",
    default=DEFAULT_CACHE_EXPIRATION,
    comment="Cache expiration in seconds",
)


def my_key_generator(namespace, func, to_str=str):  # pylint: disable=unused-argument
    """Retrieve function cache key generator"""
    if namespace is None:
        namespace = f"{func.__module__}:{func.__name__}"
    else:
        namespace = f"{func.__module__}:{func.__name__}|{namespace}"

    argspec = compat.inspect_getargspec(func)
    default_list = list(argspec.defaults or [])
    # Reverse the list, as we want to compare the argspec by negative index,
    # meaning default_list[0] should be args[-1], which works well with
    # enumerate()
    default_list.reverse()
    # use idx*-1 to create the correct right-lookup index.
    args_with_defaults = {
        argspec.args[(idx * -1)]: default for idx, default in enumerate(default_list, 1)
    }
    if argspec.args and argspec.args[0] in ("self", "cls"):
        arg_index_start = 1
    else:
        arg_index_start = 0

    def generate_key(*args, **kwargs):
        as_kwargs = {
            argspec.args[idx]: arg
            for idx, arg in enumerate(args[arg_index_start:], arg_index_start)
        }
        as_kwargs.update(kwargs)
        for arg, val in args_with_defaults.items():
            if arg not in as_kwargs:
                as_kwargs[arg] = val

        argument_values = [as_kwargs[key] for key in sorted(as_kwargs.keys()) if key != "refresh"]
        cache_key = namespace + "|" + " ".join(map(to_str, argument_values))
        log.debug(
            '<%s> generate_key(%s, %s): cache key="%s"',
            namespace,
            ", ".join(str(a) for a in args) if args else "no arg",
            ", ".join([f"{k}={v}" for k, v in kwargs.items()]) if kwargs else "no kwarg",
            cache_key,
        )
        return cache_key

    return generate_key


xdgcache = make_region(function_key_generator=my_key_generator)


def cache_on_arguments_if_not_refresh(*args, leave_refresh_kwarg=False, **kwargs):
    """
    Additional decorator method for dogpile.cache.region.CacheRegion that handle
    a special refresh parameter that permit to control cache refresh
    """
    original_decorator = xdgcache.cache_on_arguments(*args, **kwargs)

    def decorator(user_func):
        def wrapper(*args, **kwargs):
            refresh = (
                kwargs.pop("refresh", False)
                if not leave_refresh_kwarg
                else kwargs.get("refresh", False)
            )
            log.debug(
                "wrapper(%s, %s): refresh=%s, original_decorator=%s, user_func=%s",
                ", ".join([str(arg) for arg in args]),
                ", ".join([f"{key}={value}" for key, value in kwargs.items()]),
                refresh,
                original_decorator,
                user_func,
            )
            if refresh:
                log.debug("Refreshing asked: refresh cached value")
                # pylint: disable=no-member
                return original_decorator(user_func).refresh(*args, **kwargs)
            return original_decorator(user_func)(*args, **kwargs)

        return wrapper

    return decorator


xdgcache.cache_on_arguments_if_not_refresh = cache_on_arguments_if_not_refresh
