# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" SSH Client """

import errno
import fcntl
import logging
import os
import re
import select
import socket
import stat
import struct
import sys
import tempfile
import termios
import time
import traceback
import tty

from paramiko import AutoAddPolicy, SSHClient, SSHConfig, SSHException
from paramiko.agent import AgentRequestHandler
from paramiko.ssh_exception import AuthenticationException, BadAuthenticationType
from paramiko.util import ClosingContextManager

from eessh.api import api
from eessh.config import BooleanOption, IntegerOption, PasswordOption, StringOption, config
from eessh.forwarding import CForwarder, PForwarder, RForwarder, XForwarder
from eessh.utils import KB, MasterFifo, SlaveFifo, attr_data, format_size, format_time

log = logging.getLogger(__name__)

tempdir = tempfile.gettempdir()

# Register firsthop config section & options
firsthop_section = config.add_section("firsthop", comment="SSH first hop", order=2)
firsthop_section.add_option(
    BooleanOption,
    "enabled",
    default=False,
    comment="Enable/disable first hop",
    arg_help="Enable first hop",
)
firsthop_section.add_option(StringOption, "host", comment="FQDN or IP address of the first hop")
firsthop_section.add_option(IntegerOption, "port", default=22, comment="SSH port of the first hop")
firsthop_section.add_option(
    StringOption, "hostname", comment="Hostname of the first hop, for display usage only"
)
firsthop_section.add_option(
    StringOption, "username", comment="SSH username to use to connect on first hop"
)
firsthop_section.add_option(
    PasswordOption,
    "password",
    default="keyring",
    comment='SSH password to use to connect on first hop (set to "keyring" to use XDG keyring)',
    username_option="username",
    keyring_value="keyring",
)

# Register SSH connection config section & options
connection_section = config.add_section("connection", comment="SSH connection", order=3)
connection_section.add_option(
    BooleanOption,
    "agent_forwarding",
    default=False,
    comment="Enable SSH agent forwarding",
    arg="--agent-forwarding",
    short_arg="-A",
)
connection_section.add_option(
    BooleanOption,
    "compress",
    default=False,
    comment="Enable SSH compression",
    arg="--compress",
    short_arg="-C",
)
connection_section.add_option(
    BooleanOption,
    "x11_forwarding",
    default=False,
    comment="Enable X11 forwarding",
    arg="--x11-forwarding",
    short_arg="-X",
)
connection_section.add_option(
    StringOption,
    "sendenv",
    comment="List of environment variable name patterns separated by comma to send to remote",
    default="GIT_*,DEBFULLNAME,DEBEMAIL",
)
connection_section.add_option(
    BooleanOption,
    "use_ssh_config",
    default=False,
    comment="Use ssh client configuration",
)
connection_section.add_option(
    StringOption,
    "ssh_config_path",
    default="~/.ssh/config",
    comment="Ssh client configuration path",
)
connection_section.add_option(
    StringOption,
    "force_term",
    comment="Force TERM value",
)
connection_section.add_option(
    BooleanOption,
    "disable_ssh_agent_if_host_password",
    default=False,
    comment="Disable SSH agent to connect on host if we have a password for it",
    short_arg="-D",
)
connection_section.add_option(
    StringOption,
    "eecopy_download_directory",
    comment=(
        "Download directory path to store file received from eecopy transfers "
        "(default: current directory)"
    ),
)


class EeSSHClientException(Exception):
    """EeSSH Client Exception"""


class EeSSHClient:
    """EE SSH Client"""

    CONNECT_TIMEOUT = 10

    CTRL_FIFO = os.path.join(tempdir, "eessh-ctrl")
    DATA_FIFO = os.path.join(tempdir, "eessh-data")

    READ_SIZE = 1024

    def __init__(
        self,
        data,
        agent_forwarding=False,
        broadcast=False,
        compress=False,
        embedded=False,
        no_firsthop=False,
        username=None,
        x11_forwarding=False,
    ):
        self.__afwd = None
        self.__data = data
        self.__mode = None
        self.__loop = True
        self.__cfwd = None
        self.__pfwds = []
        self.__pfwds_index = 0
        self.__rfwds = []
        self.__rfwds_index = 0
        self.__sftp = None
        self.__shell = False
        self.__expect = None
        self.__xfwd = None
        self.__tty_mode = None
        self.__tty_fd = None

        self.channel = None
        self.client = None
        self.agent_forwarding = agent_forwarding
        self.compress = compress
        self.embedded = embedded
        self.no_firsthop = no_firsthop
        self.x11_forwarding = x11_forwarding

        self.stdin = None
        self.stdout = None
        self.stderr = None

        self.__alt_buffer = False
        self.__broadcast = "off"
        self.__ctrl_fifo = None
        self.__data_fifo = None

        if config.get("connection", "force_term"):
            os.environ["TERM"] = config.get("connection", "force_term")

        if username:
            # Override username and password from last hop
            data["hops"][-1]["username"] = username
            data["hops"][-1]["password"] = None
            for user in data["users"]:
                if user["username"] == username:
                    data["hops"][-1]["password"] = user["password"]
                    break

        if broadcast:
            self.do_broadcast("toggle")

    @property
    def afwd(self):
        """Check if SSH Agent request forwarding handler is opened"""
        return self.__afwd is not None

    @property
    def alt_buffer(self):
        """Check if alternative screen buffer is enabled"""
        return self.__alt_buffer

    @property
    def broadcast(self):
        """Get broadcast current mode"""
        return self.__broadcast

    @property
    def cfwd(self):
        """Check if CForwarder handler is opened"""
        return self.__cfwd is not None

    @property
    def hostname(self):
        """Get current connected server hostname"""
        return self.__data["hostname"]

    @property
    def mode(self):
        """Get current mode (shell/sftp)"""
        return self.__mode

    @property
    def pfwds(self):
        """Opened port forwarding tunnels"""
        pfwds = [
            [pfwd.index, pfwd.server_host, pfwd.server_port, pfwd.remote_host, pfwd.remote_port]
            for pfwd in self.__pfwds
        ]
        return pfwds

    @property
    def rfwds(self):
        """Opened reverse port forwarding tunnels"""
        rfwds = [
            [rfwd.index, rfwd.server_host, rfwd.server_port, rfwd.remote_host, rfwd.remote_port]
            for rfwd in self.__rfwds
        ]
        return rfwds

    @property
    def title(self):
        """Format and retrieve the window title"""
        flags = []
        if self.__broadcast == "master":
            flags.append("MASTER")

        elif self.__broadcast == "slave":
            flags.append("SLAVE")

        if self.__mode:
            flags.append(self.__mode.upper())

        if self.__afwd:
            flags.append("AFWD")

        if self.__cfwd:
            flags.append("CFWD")

        if self.__pfwds:
            flags.append("PFWD")

        if self.__rfwds:
            flags.append("RFWD")

        if self.__xfwd:
            flags.append("XFWD")

        # pylint: disable=consider-using-f-string
        title = "EeSSH - {customer_uid}{hostname}{flags}".format(
            customer_uid=(f"{self.__data['customer_uid']}-" if self.__data["customer_uid"] else ""),
            hostname=self.hostname,
            flags=f" [{'|'.join(flags)}]" if flags else "",
        )

        return title

    @property
    def xfwd(self):
        """Get X11 forwarding tunnel display"""
        if self.__xfwd is None:
            return None

        return self.__xfwd.display

    @property
    def username(self):
        """Get current connected username"""
        transport = self.channel.get_transport()
        return transport.get_username()

    def poutput(self, msg, end="\n"):
        """Helper to print message on stdout"""
        self.stdout.write(msg.encode("utf-8"))
        if end and not msg.endswith(end):
            self.stdout.write(end.encode("utf-8"))

    def perror(self, errormsg, end="\n"):
        """Helper to print error on stderr"""
        self.stderr.write(
            "\x1b[35m{}{}\x1b[39m".format(
                errormsg.replace("\n", "\r\n"), end if end and not errormsg.endswith(end) else ""
            ).encode("utf-8")
        )

    def __process(self):
        """Start processing shell event"""
        buf = b""

        def process_data(data):
            length = len(data)
            if length == 1 and not self.embedded:
                value = ord(data)
                if value == 0:
                    # CTRL+SPACE
                    if self.__broadcast == "slave":
                        return

                    self.open_eeshell()
                    return

                if value == 2:
                    # CTRL+B
                    self.do_broadcast("toggle")
                    return

                if value == 21:
                    # CTRL+U
                    self.do_su()
                    return

            if self.__mode == "shell":
                # Send data to remote
                if self.channel.send(data) != length:
                    self.__loop = False

                if self.__broadcast == "master":
                    self.__data_fifo.send(data)

        term_size = None
        while self.__loop:
            # Resize the terminal if needed
            size = self.get_terminal_size()
            if size != term_size:
                self.channel.resize_pty(width=size[1], height=size[0])
                term_size = size

            # Handle input/output
            read_list = [self.channel, self.stdin]

            if self.__broadcast == "slave":
                read_list.append(self.__ctrl_fifo)
                read_list.append(self.__data_fifo)

            try:
                # pylint: disable=unused-variable
                rlist, wlist, xlist = select.select(read_list, [], [], 0.1)

            except OSError as err:
                if err.errno == errno.EINTR:
                    # May happen when resizing terminal
                    continue
                raise

            if self.channel in rlist:
                data = self.channel.recv(self.READ_SIZE)

                if data == CForwarder.CMD_START:
                    # Copy detection
                    if self.open_cfwd():
                        self.channel.send(
                            CForwarder.CMD_READY
                            + b":"
                            + bytes(str(self.__cfwd.server_port), "utf-8")
                            + b"\n"
                        )

                elif data == CForwarder.CMD_CLOSE:
                    # Copy detection
                    if self.__cfwd:
                        self.close_cfwd()
                    self.channel.send(CForwarder.CMD_CLOSE + b"\n")

                elif self.__mode == "shell":
                    self.stdout.write(data)
                    self.stdout.flush()

                    if self.__expect:
                        buf += data
                        pattern = self.__expect.get("pattern")
                        if pattern and pattern.search(buf):
                            response = self.__expect.get("response")
                            if response:
                                log.debug("[EXPECT] Buffer matched pattern, sending response")
                                self.channel.send(response)
                            else:
                                log.warning(
                                    "[EXPECT] Buffer '%s' NOT matched with pattern %s, expect failed",
                                    pattern,
                                    buf,
                                )

                            buf = b""
                            self.__expect = None

            if self.stdin in rlist:
                data = self.stdin.read(self.READ_SIZE)
                process_data(data)

            if self.__broadcast == "slave":
                if self.__data_fifo in rlist:
                    data = self.__data_fifo.recv(self.READ_SIZE)
                    process_data(data)

                if self.__ctrl_fifo in rlist:
                    cmd = self.__ctrl_fifo.recv(self.READ_SIZE).decode("utf-8").rstrip("\n")
                    if cmd == "broadcast_off":
                        self.do_broadcast("off")
                    elif cmd == "do_su":
                        self.do_su()
                    else:
                        log.error("[SLAVE] received an invalid command %s", cmd)

            if self.__mode == "shell" and self.channel.exit_status_ready():
                self.__loop = False

    def copy(self):
        """Return a copy of this client"""

        broadcast = self.__broadcast == "off"

        client = EeSSHClient(
            self.__data.copy(),
            agent_forwarding=self.agent_forwarding,
            broadcast=broadcast,
            compress=self.compress,
            embedded=self.embedded,
            x11_forwarding=self.x11_forwarding,
        )

        return client

    def do_broadcast(self, broadcast):
        """Set and start broadcast mode"""
        assert broadcast in (
            "master",
            "off",
            "slave",
            "toggle",
        ), f"Invalid broadcast mode {broadcast}"

        if broadcast == "toggle":
            if self.__broadcast in ("master", "slave"):
                broadcast = "off"
            elif not os.path.exists(self.CTRL_FIFO):
                broadcast = "master"
            else:
                broadcast = "slave"

        elif broadcast == self.__broadcast:
            return True

        log.debug("[BROADCAST] changing broadcast mode from %s to %s", self.__broadcast, broadcast)

        if self.__ctrl_fifo:
            if self.__broadcast == "master":
                # Turning slaves off
                self.__ctrl_fifo.send(b"broadcast_off\n")

            self.__ctrl_fifo.close()
            self.__ctrl_fifo = None

        if self.__data_fifo:
            self.__data_fifo.close()
            self.__data_fifo = None

        self.__broadcast = "off"

        if broadcast == "off":
            self.set_title()
            return True

        class_ = MasterFifo if broadcast == "master" else SlaveFifo

        def create_fifo(filepath, class_=class_):
            try:
                fifo = class_(filepath)
                fifo.start()

            except Exception:  # pylint: disable=broad-except
                log.exception("Failed to create fifo file")
                sys.stderr.write("Failed to create fifo file\n")
                return None

            return fifo

        self.__ctrl_fifo = create_fifo(self.CTRL_FIFO)
        if self.__ctrl_fifo is None:
            return False

        self.__data_fifo = create_fifo(self.DATA_FIFO)
        if self.__data_fifo is None:
            self.__ctrl_fifo.close()
            self.__ctrl_fifo = None
            return False

        self.__broadcast = broadcast

        self.set_title()
        return True

    def do_cat(self, path):
        """Cat a file on remote"""
        assert self.__sftp, "Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_cat()"

        try:
            path_stat = self.__sftp.stat(path)
            size = path_stat.st_size
            if size > 512 * KB:
                self.perror("File size exceed maximum authorized (512KiB).")
                return None

            with self.__sftp.open(path, mode="r", bufsize=1024) as file_desc:
                content = file_desc.read()

        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to open file")
            return None

        return content

    def do_cd(self, path):
        """Change directory on current SFTP connection"""
        assert self.__sftp, "Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_cd()"

        try:
            self.__sftp.chdir(path)

        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to change directory")
            return False

        return True

    def do_get(self, remotepath, localpath=None):
        """Get a file on remote using current SFTP connection"""
        assert self.__sftp, "Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_get()"

        filename = os.path.basename(remotepath)

        if localpath is None:
            localpath = filename

        elif os.path.isdir(localpath):
            localpath = os.path.join(localpath, filename)

        counts = {
            "s": time.time(),
            "ds": time.time(),
            "transferred": 0,
        }

        try:
            self.__sftp.get(
                remotepath,
                localpath,
                callback=lambda transferred, total: self.sftp_callback(
                    filename, transferred, total, counts
                ),
            )

        except Exception:  # pylint: disable=broad-except
            log.exception(
                "[SFTP] Transfer failed (remotepath=%s, localpath=%s)", remotepath, localpath
            )
            return False

        finally:
            self.poutput("\n")

        return True

    def do_ls(self, path):
        """List files on specified remote path using current SFTP connection"""
        assert self.__sftp, "Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_ls()"

        try:
            attr = self.__sftp.stat(path)
            is_dir = stat.S_IFMT(attr.st_mode) == stat.S_IFDIR
            data = self.__sftp.listdir_attr(path) if is_dir else [attr]
        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to list %s", path)
            return None

        result = []
        for attr in data:
            attr = attr_data(attr)
            attr["path"] = path if not is_dir else os.path.join(path, attr["filename"])
            if not is_dir:
                attr["filename"] = os.path.basename(path)
            if attr["type"] == "l":
                attr["link_attr"] = attr_data(self.__sftp.stat(attr["path"]))
                attr["link_attr"]["path"] = self.__sftp.readlink(attr["path"])

            result.append(attr)

        return result

    def do_put(self, localpath, remotepath=None):
        """Put a local file on remote using current SFTP connection"""
        assert self.__sftp, "Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_put()"

        filename = os.path.basename(localpath)

        if remotepath is None:
            remotepath = filename

        else:
            try:
                path_stat = self.__sftp.stat(remotepath)
                if stat.S_ISDIR(path_stat.st_mode):
                    remotepath = os.path.join(remotepath, filename)

            # pylint: disable=broad-except
            except Exception:  # nosec
                pass

        counts = {
            "s": time.time(),
            "ds": time.time(),
            "transferred": 0,
        }

        try:
            self.__sftp.put(
                localpath,
                remotepath,
                callback=lambda transferred, total: self.sftp_callback(
                    filename, transferred, total, counts
                ),
            )

        except Exception:  # pylint: disable=broad-except
            log.exception(
                "[SFTP] Transfer failed (localpath=%s, remotepath=%s)", localpath, remotepath
            )
            return False

        finally:
            self.poutput("\n")

        return True

    def do_pwd(self):
        """Retrieve current SFTP connection working directory path"""
        assert self.__sftp, "Must call EeSSHClient.open_sftp() before calling EeSSHClient.do_pwd()"

        return self.__sftp.getcwd() or "~/"

    def do_su(self):
        """Switch as root user on remote"""
        # Get parameters from last hop
        data = self.__data["hops"][-1]

        su_method = data.get("su_method")
        if su_method is None:
            self.perror("\n\rMissing super user method.\n\r", end=None)
            self.channel.send(b"\n")
            return

        if su_method == "enable":
            command = "enable"
            pattern = rb"^(Password|Not de passe)\s*:\s*$"
            response = data.get("su_password")

        elif su_method == "su":
            # Make sure to disable bracketed-paste
            # (enabled by default in Bash >= 5.1)
            self.channel.send(b'bind "set enable-bracketed-paste off"\n')
            command = "su -"
            pattern = rb"^(Password|Not de passe)\s*:\s*$"
            response = data.get("su_password")

        elif su_method == "sudo":
            # Make sure to disable bracketed-paste
            # (enabled by default in Bash >= 5.1)
            self.channel.send(b'bind "set enable-bracketed-paste off"\n')
            command = "sudo -i -p 'SUDO PASSWORD:'"
            pattern = rb"^SUDO PASSWORD:$"
            response = data.get("su_password") or data.get("password")

        else:
            self.perror(f"Invalid root method {su_method}")
            return

        self.__expect = {
            "pattern": re.compile(pattern, re.M),
            "response": (f"{response}\n".encode() if response else None),
        }

        self.channel.send(f"{command}\n".encode())

        if self.__broadcast == "master":
            self.__ctrl_fifo.send(b"do_su\n")

    def close(self):
        """Close client"""
        self.do_broadcast("off")

        if self.__afwd:
            # Closing agent forwarding tunnel
            self.__afwd.close()
            self.__afwd = None

        if self.__cfwd:
            # Closing copy tunnel
            self.__cfwd.close()
            self.__cfwd = None

        if self.__xfwd:
            # Closing X11 forwarding tunnel
            self.__xfwd.close()
            self.__xfwd = None

        # Closing port forwarding tunnels
        if self.__pfwds:
            self.close_pfwd("all")

        # Closing reverse port forwarding tunnels
        if self.__rfwds:
            self.close_rfwd("all")

        if self.channel:
            self.channel.close()

        # Reset window title
        self.set_title(reset=True)

    def close_cfwd(self):
        """Close opened copy tunnel with remote"""

        if self.__cfwd is None:
            self.perror("Copy tunnel not opened.")
            return True

        self.poutput(f"Closing {self.__cfwd.name}")

        result = self.__cfwd.close()
        self.__cfwd = None

        self.set_title()
        return result

    def close_pfwd(self, index):
        """Close opened port forwarding tunnel with remote"""

        if not self.__pfwds:
            self.perror("No port forwarding tunnel opened.")
            return True

        if index == "all":
            for pfwd in self.__pfwds:
                self.poutput(
                    f"Closing port forwarding tunnel from {pfwd.server_host}:{pfwd.server_port} "
                    f"(on local) to {pfwd.remote_host}:{pfwd.remote_port} (on remote)"
                )
                pfwd.close()

            self.__pfwds = []
            self.set_title()
            return True

        found = False
        for pfwd in self.__pfwds:
            if str(pfwd.index) == str(index):
                found = True
                break

        if not found:
            self.perror(f"Port forwarding tunnel #{index} not found.")
            return True

        self.poutput(
            f"Closing port forwarding tunnel from {pfwd.server_host}:{pfwd.server_port} (on "
            f"local) to {pfwd.remote_host}:{pfwd.remote_port} (on remote)"
        )

        result = pfwd.close()

        self.__pfwds.remove(pfwd)
        self.set_title()
        return result

    def close_rfwd(self, index):
        """Close opened reverse port forwarding tunnel with remote"""

        if not self.__rfwds:
            self.perror("No reverse port forwarding tunnel opened.")
            return True

        if index == "all":
            for rfwd in self.__rfwds:
                self.poutput(
                    f"Closing reverse port forwarding tunnel from {rfwd.server_host}:"
                    f"{rfwd.server_port} (on remote) to {rfwd.remote_host}:{rfwd.remote_port} "
                    "(on local)"
                )
                rfwd.close()

            self.__rfwds = []
            self.set_title()
            return True

        found = False
        for rfwd in self.__rfwds:
            if str(rfwd.index) == str(index):
                found = True
                break

        if not found:
            self.perror(f"Reverse port forwarding tunnel #{index} not found.")
            return True

        self.poutput(
            f"Closing reverse port forwarding tunnel from {rfwd.server_host}:{rfwd.server_port} "
            f"(on remote) to {rfwd.remote_host}:{rfwd.remote_port} (on local)"
        )

        result = rfwd.close()

        self.__rfwds.remove(rfwd)
        self.set_title()
        return result

    def close_sftp(self):
        """Close opened SFTP channel with remote"""

        if self.__sftp is None:
            return True

        self.__sftp.close()
        self.__sftp = None
        self.set_title()

        return True

    def connect(self):
        """Connect on remote"""
        if self.client:
            return True

        hops = self.__data["hops"][:]
        if not hops:
            self.stderr.write(
                b"\r[\x1b[31mFAILED\x1b[39m]\n no connection information retrieved for server.\n\r"
            )
            return False

        if config.get("firsthop", "enabled") and not self.no_firsthop:
            if not config.get("firsthop", "host") or not config.get("firsthop", "username"):
                self.stderr.write(
                    b"\r[\x1b[31mFAILED\x1b[39m] firsthop host miss-configured: you have to set "
                    b"at least host and username.\n\r"
                )
                return False
            hops.insert(
                0,
                {
                    "host": config.get("firsthop", "host"),
                    "hostname": (
                        config.get("firsthop", "hostname")
                        if config.get("firsthop", "hostname")
                        else config.get("firsthop", "host")
                    ),
                    "port": config.get("firsthop", "port"),
                    "username": config.get("firsthop", "username"),
                    "password": config.get("firsthop", "password"),
                },
            )

        sock = None
        previous_hop = None
        data = None
        ssh_config = None
        if config.get("connection", "use_ssh_config"):
            ssh_config = SSHConfig()
            ssh_config_path = os.path.expanduser(config.get("connection", "ssh_config_path"))
            self.stdout.write(
                bytes(f"[ .... ] Loading SSH configuration from {ssh_config_path}", "utf-8")
            )
            try:
                with open(ssh_config_path, encoding="utf8") as ssh_config_fd:
                    ssh_config.parse(ssh_config_fd)
            except Exception:  # pylint: disable=broad-except
                self.stderr.write(b"\r[\x1b[31mFAILED\x1b[39m]\n exiting...\n\r")
                self.perror(traceback.format_exc())
                log.error("Failed to load SSH configuration from %s", ssh_config_path)
                return False
            self.stdout.write(b"\r[  \x1b[32mOK\x1b[39m  ]\n\r")

        for index, data in enumerate(hops):
            if index > 0:
                try:
                    self.stdout.write(
                        bytes(
                            # pylint: disable=consider-using-f-string
                            "[ .... ] Start NC proxy to {hostname}:{port} on {previous_hop}".format(
                                # pylint: disable=unsubscriptable-object
                                previous_hop=previous_hop["hostname"],
                                **data,
                            ),
                            "utf-8",
                        )
                    )  # pylint: disable=unsubscriptable-object
                    sock = NCProxy(self.client)
                    sock.connect(data["host"], port=data["port"])
                    self.stdout.write(b"\r[  \x1b[32mOK\x1b[39m  ]\n\r")
                except Exception:  # pylint: disable=broad-except
                    self.stderr.write(b"\r[\x1b[31mFAILED\x1b[39m]\n exiting...\n\r")
                    self.perror(traceback.format_exc())
                    log.error(
                        "Failed to start NC proxy to %s:%s on %s",
                        data["host"],
                        data["port"],
                        previous_hop["hostname"],  # pylint: disable=unsubscriptable-object
                    )
                    return False

            # Enable compression for last hop if requested
            compress = self.compress and index == len(self.__data["hops"]) - 1

            self.stdout.write(
                # pylint: disable=consider-using-f-string
                bytes("[ .... ] Connecting to {username}@{hostname}:{port}".format(**data), "utf-8")
            )

            self.client = SSHClient()
            self.client.set_missing_host_key_policy(AutoAddPolicy())

            host_ssh_config = ssh_config.lookup(data["host"]) if ssh_config else {}
            log.debug("Host %s SSH configuration: %s", data["host"], host_ssh_config)
            connect_kwargs = {
                "port": host_ssh_config.get("port", data["port"]),
                "sock": sock,
                "username": host_ssh_config.get("user", data["username"]),
                "password": data.get("password"),
                "key_filename": host_ssh_config.get("identityfile", None),
                "compress": compress,
                "timeout": self.CONNECT_TIMEOUT,
            }
            if (
                config.get("connection", "disable_ssh_agent_if_host_password")
                and connect_kwargs["password"]
                and connect_kwargs["username"] == data["username"]
            ):
                log.info(
                    "Disable SSH agent for connection on %s@%s (force use password)",
                    data["username"],
                    data["hostname"],
                )
                connect_kwargs["allow_agent"] = False
            try:
                self.client.connect(data["host"], **connect_kwargs)
            except (BadAuthenticationType, AuthenticationException) as err:
                log.debug(
                    "%s exception occurred connecting to %s, retry without "
                    "'rsa-sha2-512' and 'rsa-sha2-256' pubkeys algorithms",
                    type(err).__name__,
                    data["host"],
                )
                self.stderr.write(
                    b"\r[\x1b[31mFAILED\x1b[39m]\n Authentication failed, retry without "
                    b"'rsa-sha2-512' and 'rsa-sha2-256' pubkeys algorithms...\n\r"
                )
                if sock:
                    try:
                        # We need to restart the NCproxy, before retrying to connect
                        self.stdout.write(
                            bytes(
                                # pylint: disable=consider-using-f-string
                                "[ .... ] Restart NC proxy to {hostname}:{port} on {previous_hop}".format(
                                    # pylint: disable=unsubscriptable-object
                                    previous_hop=previous_hop["hostname"],
                                    **data,
                                ),
                                "utf-8",
                            )
                        )  # pylint: disable=unsubscriptable-object
                        sock.restart()
                        self.stdout.write(b"\r[  \x1b[32mOK\x1b[39m  ]\n\r")
                    except Exception:  # pylint: disable=broad-except
                        self.stderr.write(b"\r[\x1b[31mFAILED\x1b[39m]\n exiting...\n\r")
                        self.perror(traceback.format_exc())
                        log.error(
                            "Failed to restart NC proxy to %s:%s on %s",
                            data["host"],
                            data["port"],
                            previous_hop["hostname"],  # pylint: disable=unsubscriptable-object
                        )
                        return False
                self.stdout.write(
                    # pylint: disable=consider-using-f-string
                    bytes(
                        "[ .... ] Connecting to {username}@{hostname}:{port} without "
                        "'rsa-sha2-512' and 'rsa-sha2-256' pubkeys algorithms".format(**data),
                        "utf-8",
                    )
                )
                try:
                    connect_kwargs["disabled_algorithms"] = {
                        "pubkeys": ["rsa-sha2-512", "rsa-sha2-256"]
                    }
                    self.client.connect(data["host"], **connect_kwargs)
                except Exception:  # pylint: disable=broad-except
                    self.client = None
                    self.stderr.write(b"\r[\x1b[31mFAILED\x1b[39m]\n exiting...\n\r")
                    self.perror(traceback.format_exc())
                    log.error(
                        "Failed to connect to %s even without 'rsa-sha2-512' and 'rsa-sha2-256' "
                        "pubkeys algorithms",
                        data["hostname"],
                    )
                    return False
            except Exception:  # pylint: disable=broad-except
                self.client = None
                self.stderr.write(b"\r[\x1b[31mFAILED\x1b[39m]\n exiting...\n\r")
                self.perror(traceback.format_exc())
                log.error("Failed to connect to %s", data["hostname"])
                return False

            self.stdout.write(b"\r[  \x1b[32mOK\x1b[39m  ]\n\r")
            previous_hop = data

        self.poutput("\n\n")

        # Open session
        self.channel = self.client.get_transport().open_session()

        # Send environment variables
        sendenv_patterns = [
            re.compile("^" + pattern.replace("*", ".+").strip() + "$")
            for pattern in config.get("connection", "sendenv").split(",")
        ]
        log.debug(
            "Send environment variable name patterns: %s",
            ", ".join([str(sendenv_pattern) for sendenv_pattern in sendenv_patterns]),
        )
        for var_name, var_value in os.environ.items():
            send = False
            for pattern in sendenv_patterns:
                if pattern.match(var_name):
                    send = True
                    break
            if send:
                log.debug('Send environment variable "%s": "%s"', var_name, var_value)
                self.channel.set_environment_variable(var_name, var_value)

        # Opening agent forwarding tunnel if needed
        if self.agent_forwarding:
            self.open_afwd()

        # Opening X11 forwarding tunnel if needed
        if self.x11_forwarding:
            self.open_xfwd()

        # Set default path
        data["path"] = "~"

        return True

    def get_terminal_size(self):
        """Retrieve terminal size"""
        return struct.unpack("hh", fcntl.ioctl(self.__tty_fd, termios.TIOCGWINSZ, "1234"))

    def get_terminal_height(self):
        """Retrieve terminal height"""
        size = self.get_terminal_size()
        return size[0]

    def get_terminal_width(self):
        """Retrieve terminal width"""
        size = self.get_terminal_size()
        return size[1]

    def get_server_data(self, refresh=False):
        """Get current connected server extended info using API"""
        if not self.__data["customer_uid"] or not self.__data["hostname"]:
            return False
        return api.get_data(self.__data["customer_uid"], self.__data["hostname"], refresh=refresh)

    def get_server_tickets(self, refresh=False):
        """Get current connected server tickets using API"""
        if not self.__data["customer_uid"] or not self.__data["hostname"]:
            return False
        return api.get_tickets(
            self.__data["customer_uid"], self.__data["hostname"], refresh=refresh
        )

    def update_data(self, changes):
        """Update current connected server info using API"""
        if not self.__data["customer_uid"] or not self.__data["hostname"]:
            return False
        return api.update_data(
            customer_uid=self.__data["customer_uid"],
            hostname=self.__data["hostname"],
            changes=changes,
        )

    def open_afwd(self):
        """Open SSH Agent request forwarding handler"""

        if self.__afwd:
            self.perror("SSH Agent request forwarding handler is already opened.\n\n\r", end=None)
            return False

        if self.__shell:
            self.perror("A shell is already invoked.\n\n\r", end=None)
            return False

        self.__afwd = AgentRequestHandler(self.channel)

        self.set_title()
        return True

    def open_cfwd(self):
        """Open copy tunnel with remote"""

        if self.__cfwd:
            self.perror("Copy tunnel is already opened.\n\n\r", end=None)
            return True

        try:
            cfwd = CForwarder(
                self.client, dirpath=config.get("connection", "eecopy_download_directory")
            )

        except Exception:  # pylint: disable=broad-except
            self.perror("Failed to copy tunnel")
            return False

        if not cfwd.open():
            self.perror(f"Failed to open {cfwd.name}")
            return False

        self.poutput(f"Opening {cfwd.name}")

        self.__cfwd = cfwd

        self.set_title()
        return True

    def open_eeshell(self):
        """Open EEShell"""

        from eessh.shell import EeCMDShell  # pylint: disable=import-outside-toplevel

        # Enable alternative screen buffer
        self.switch_buffer()

        self.open_sftp()

        cmd = EeCMDShell(self)
        lastcmd = cmd.run()

        self.close_sftp()

        # Disable alternative screen buffer
        self.switch_buffer()

        self.open_shell()

        if lastcmd == "su":
            self.do_su()

    def open_pfwd(self, server_host, server_port, remote_host, remote_port):
        """Open a port forwarding tunnel with remote"""

        try:
            self.__pfwds_index += 1
            pfwd = PForwarder(
                self.client,
                server_host,
                server_port,
                remote_host,
                remote_port,
                index=self.__pfwds_index,
            )

        except Exception:  # pylint: disable=broad-except
            self.perror(
                f"Failed to open port forwarding tunnel from {server_host}:{server_port} (on "
                f"local) to {remote_host}:{remote_port} (on remote)"
            )
            return False

        if not pfwd.open():
            self.perror(
                f"Failed to open port forwarding tunnel from {server_host}:{server_port} (on "
                f"local) to {remote_host}:{remote_port} (on remote)"
            )
            return False

        self.poutput(
            f"Opening port forwarding tunnel from {server_host}:{server_port} (on local) to "
            f"{remote_host}:{remote_port} (on remote)"
        )

        self.__pfwds.append(pfwd)

        self.set_title()
        return True

    def open_rfwd(self, server_host, server_port, remote_host, remote_port):
        """Open a reverse port forwarding tunnel with remote"""

        try:
            self.__rfwds_index += 1
            rfwd = RForwarder(
                self.client,
                server_host,
                server_port,
                remote_host,
                remote_port,
                index=self.__rfwds_index,
            )

        except Exception:  # pylint: disable=broad-except
            self.perror(
                f"Failed to open reverse port forwarding tunnel from {server_host}:{server_port} "
                f"(on remote) to {remote_host}:{remote_port} (on local)"
            )
            return False

        if not rfwd.open():
            self.perror(
                f"Failed to open reverse port forwarding tunnel from {server_host}:{server_port} "
                f"(on remote) to {remote_host}:{remote_port} (on local)"
            )
            return False

        self.poutput(
            f"Opening reverse port forwarding tunnel from {server_host}:{server_port} (on "
            f"remote) to {remote_host}:{remote_port} (on local)"
        )

        self.__rfwds.append(rfwd)

        self.set_title()
        return True

    def open_sftp(self):
        """Open SFTP channel with remote"""

        if self.__sftp:
            self.__mode = "sftp"
            return True

        if not self.connect():
            return False

        try:
            self.__sftp = self.client.open_sftp()

        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to open SFTP")
            return False

        self.__mode = "sftp"
        self.set_title()
        return True

    def open_shell(self):
        """Open shell on remote"""

        if self.__shell:
            self.__mode = "shell"
            self.set_title()
            return True

        if not self.connect():
            return False

        # Invoke shell on server
        try:
            size = self.get_terminal_size()

            self.channel.get_pty(
                term=os.environ.get("TERM", "xterm"), width=size[1], height=size[0]
            )
            self.channel.invoke_shell()
            self.__shell = True

        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to invoke shell on remote")
            return False

        self.__mode = "shell"
        self.set_title()
        return True

    def open_xfwd(self):
        """Open a X11 forwarding tunnel with remote"""

        if self.__xfwd:
            self.perror("X11 forwarding tunnel is already opened.\n\n\r", end=None)
            return False

        if self.__shell:
            self.perror("A shell is already invoked.\n\n\r", end=None)
            return False

        x11_display = os.environ.get("DISPLAY")
        if x11_display is None:
            self.perror("Failed to get X11 display, disabling X11 forwarding\n\n\r", end=None)
            return False

        try:
            self.__xfwd = XForwarder(self.channel, x11_display)

        except Exception:  # pylint: disable=broad-except
            self.perror(f"Failed to open X11 forwarding tunnel to {x11_display}\n\n\r", end=None)
            return False

        if not self.__xfwd.open():
            self.perror(f"Failed to open X11 forwarding tunnel to {x11_display}\n\n\r", end=None)
            self.__xfwd = None
            return False

        self.stdout.write(bytes(f"Opening X11 forwarding tunnel to {x11_display}\n\n\r", "utf-8"))

        self.set_title()
        return True

    def prepare_tty(self):
        """
        Disable echo local and set TTY to raw mode so that stdin is ready even if enter is not pressed
        """

        # Disable echo
        mode = termios.tcgetattr(self.__tty_fd)
        mode[tty.LFLAG] &= ~termios.ECHO

        termios.tcsetattr(self.__tty_fd, termios.TCSADRAIN, mode)

        # Set raw mode
        tty.setraw(self.__tty_fd)

    def restore_tty(self):
        """
        Enable echo local and disable raw mode
        """

        termios.tcsetattr(self.__tty_fd, termios.TCSADRAIN, self.__tty_mode)

    def run(self, action="shell", on_ready=None, on_exit=None, tty_fd=None):
        """Run the client"""
        # Reopen standard input/output/error in binary mode
        self.stdin = os.fdopen(sys.stdin.fileno() if tty_fd is None else tty_fd, "rb", 0)
        self.stdout = os.fdopen(sys.stdout.fileno() if tty_fd is None else tty_fd, "wb", 0)
        self.stderr = os.fdopen(sys.stderr.fileno() if tty_fd is None else tty_fd, "wb", 0)

        # Save current TTY mode for later restoration
        self.__tty_fd = self.stdout.fileno()
        self.__tty_mode = termios.tcgetattr(self.__tty_fd)

        self.prepare_tty()

        if not self.connect():
            self.close()
            self.restore_tty()
            if on_exit:
                on_exit(self, False)
            return False

        if action == "shell":
            self.open_shell()

        elif action == "eeshell":
            self.open_eeshell()

        else:
            # Invalid action
            self.close()
            self.restore_tty()
            if on_exit:
                on_exit(self, False)
            return False

        self.set_title()

        if on_ready:
            on_ready(self)

        result = True
        try:
            self.__process()
        except Exception:  # pylint: disable=broad-except
            log.exception("An exception occurred during communication with remote")
            result = False

        self.close()
        self.restore_tty()

        if on_exit:
            on_exit(self, result)

        return result

    def set_title(self, reset=False):
        """
        Change window's title
        """

        if self.stdout is None:
            return False

        self.stdout.write(f"\033]0;{'' if reset else self.title}\007".encode())

        return True

    def sftp_callback(self, filename, transferred, total, counts):
        """
        SFTP callback to display useful information during file transfer
        """

        current_time = time.time()
        if current_time - counts["ds"] >= 1:
            chunk = transferred - counts["transferred"]
            speed = chunk // (current_time - counts["ds"])
            percent = transferred * 100 // total
            remaining = int(
                (total * (current_time - counts["s"]) / transferred) - (current_time - counts["s"])
            )

            width = self.get_terminal_width()
            txt_r = f"{percent:3d}% {format_size(transferred)} {format_size(speed)}/s {format_time(remaining)}"
            txt_l = f"{filename:{width - len(txt_r) - 1}s}"

            self.stdout.write(bytes(f"\r{txt_l} {txt_r}", "utf-8"))

            counts["ds"] = current_time
            counts["transferred"] = transferred

        if transferred == total:
            speed = total // (current_time - counts["s"])

            width = self.get_terminal_width()
            percent = 100
            duration = int(current_time - counts["s"])

            txt_r = f"{percent:3d}% {format_size(transferred)} {format_size(speed)}/s {format_time(duration)}"
            txt_l = f"{filename:{width - len(txt_r) - 1}s}"

            self.stdout.write(bytes(f"\r{txt_l} {txt_r}", "utf-8"))

    def stop(self):
        """Stop the client"""
        self.__loop = False

    def switch_buffer(self):
        """Switch active buffer"""

        if not self.__alt_buffer:
            # Enable alternative screen buffer
            self.restore_tty()
            self.stdout.write(b"\x1b[?1049h\x1b[H")

            self.__alt_buffer = True
            return True

        # Disable alternative screen buffer
        self.stdout.write(b"\x1b[?1049l")
        self.prepare_tty()

        self.__alt_buffer = False
        return True


class NCProxy(ClosingContextManager):
    """
    See `paramiko.proxy.ProxyCommand` for original implementation.
    """

    def __init__(self, client):
        self.client = client
        self.host = None
        self.port = 22

        self.stdin = None
        self.stdout = None
        self.stderr = None

        self.timeout = None

    @property
    def _closed(self):
        """Check if proxy is closed (expected by Paramiko)"""
        return self.closed

    @property
    def closed(self):
        """Check if proxy is closed"""
        if self.stdin is None:
            return True

        return self.stdin.closed

    def close(self):
        """Close the proxy"""
        if self.stdin is None:
            return True

        self.stdin.close()
        self.stdin = None

        if self.stdout is not None:
            self.stdout.close()
            self.stdout = None

        if self.stderr is not None:
            self.stderr.close()
            self.stderr = None

        return True

    def connect(self, host, port=None):
        """Connect on remote"""
        self.host = host
        if port:
            self.port = port

        log.info("[PROXY] [%s:%s] Starting NC", self.host, self.port)

        # Start NC on client
        try:
            stdin, stdout, stderr = self.client.exec_command(  # nosec
                f"nc -q0 {self.host} {self.port}"
            )
        except SSHException:
            log.error("[PROXY] [%s:%s] Exception occurred starting NC", self.host, self.port)
            raise

        # Wait a bit for NC command to be started and/or to detect any error
        time.sleep(0.1)
        if stdout.channel.exit_status_ready() and stdout.channel.recv_exit_status() != 0:
            if b"nc: invalid option -- q" in stderr.read():
                log.debug(
                    "GNU netcat detected, retry without -q parameter (no default timeout with GNU netcat)"
                )
                try:
                    stdin, stdout, stderr = self.client.exec_command(  # nosec
                        f"nc {self.host} {self.port}"
                    )
                except SSHException:
                    log.error(
                        "[PROXY] [%s:%s] Exception occurred starting GNU NC", self.host, self.port
                    )
                    raise
                if stdout.channel.exit_status_ready() and stdout.channel.recv_exit_status() != 0:
                    raise EeSSHClientException(
                        f"[PROXY] [{self.host}:{self.port}] Fail to start GNU NC: {stderr.read()}"
                    )
            else:
                raise EeSSHClientException(
                    f"[PROXY] [{self.host}:{self.port}] Fail to start NC: {stderr.read()}"
                )

        # Wait a bit for NC command to be started and/or to detect any error
        time.sleep(0.1)

        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr

    def restart(self):
        """Restart proxying"""
        self.close()
        self.connect(self.host, port=self.port)

    def send(self, content):
        """Send content via the proxy host"""
        try:
            self.stdin.write(content)

        except OSError:
            log.exception("[PROXY] [%s:%s] IOError while sending data.", self.host, self.port)
            raise

        return len(content)

    def recv(self, size):
        """Receive content via the proxy host"""
        try:
            buf = b""
            while len(buf) < size:
                if self.stdout is None:
                    break

                # pylint: disable=unused-variable
                rlist, wlist, xlist = select.select([self.stdout.channel], [], [], self.timeout)
                if rlist and self.stdout.channel is rlist[0]:
                    buf += self.stdout.read(size - len(buf))

            return buf

        except socket.timeout:
            if buf:
                return buf
            raise

        except OSError:
            log.exception("[PROXY] [%s:%s] IOError while receiving data.", self.host, self.port)
            raise

    def settimeout(self, timeout):
        """Set the proxy timeout"""
        self.timeout = timeout
