#!/usr/bin/env python
# pylint: disable=invalid-name,consider-using-f-string

# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Copy script """


import argparse
import fcntl
import getpass
import hashlib
import os.path
import re
import select
import signal
import socket
import struct
import sys
import termios
import time

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

CHUNK_SIZE = 8192
TIMEOUT = 10

CMD_ABORT = "eessh-copy-abort"
CMD_CLOSE = "eessh-copy-close"
CMD_READY = re.compile(r"eessh-copy-ready:(\d+)")
CMD_START = "eessh-copy-start"

CMD_NAME = "eessh-copy-name"
CMD_SIZE = "eessh-copy-size"
CMD_HASH = "eessh-copy-hash"
CMD_DONE = "eessh-copy-done"

STAGE_INIT = 0
STAGE_NAME = 1
STAGE_SIZE = 3
STAGE_HEAD = 5
STAGE_DATA = 6
STAGE_HASH = 8
STAGE_DONE = 10

KB = 1024
MB = 1024 * KB
GB = 1024 * MB
TB = 1024 * GB


# Python2 compat
if PY3:

    def to_bytes(string):
        """Python3 encoded string helper"""
        return bytes(string, encoding="utf-8")

else:

    def to_bytes(string):
        """Python2 encoded string helper"""
        return string


class Timeout(Exception):
    """Timeout exception class"""


def read_input(prompt, timeout):
    """Read input data"""

    # set signal handler
    def alarm_handler(signum, frame):
        raise Timeout()

    # produce SIGALRM in `timeout` seconds
    signal.signal(signal.SIGALRM, alarm_handler)
    signal.alarm(timeout)

    # Waiting for buffer to be flushed
    time.sleep(0.1)

    try:
        return getpass.getpass(prompt=prompt)
    except Timeout:
        pass
    finally:
        signal.alarm(0)

    return None


def copyfile(filepath):
    """Copy a file"""
    if not os.path.isfile(filepath):
        sys.stderr.write("Invalid file %s\n" % filepath)
        sys.stderr.flush()
        return False

    if not os.access(filepath, os.R_OK):
        sys.stderr.write("Not enough permissions to access file %s\n" % filepath)
        sys.stderr.flush()
        return False

    # Get terminal size
    # pylint: disable=unused-variable
    term_height, term_width = struct.unpack(
        "hh", fcntl.ioctl(sys.stdout, termios.TIOCGWINSZ, "1234")
    )

    filehash = hashlib.sha256()
    filename = os.path.basename(filepath)
    filesize = os.path.getsize(filepath)

    sys.stdout.write("Initializing...\n")
    sys.stdout.flush()

    result = read_input(CMD_START, TIMEOUT)
    if result is None:
        sys.stderr.write("\rTimeout, exiting.\n")
        sys.stderr.flush()
        return False

    # Wait for remote to open tunnel
    match = CMD_READY.match(result)
    if match is None:
        sys.stderr.write("\nHandshake failed, exiting.\n")
        sys.stderr.flush()
        return False

    port = int(match.group(1))

    sock = socket.socket()
    try:
        sock.connect(("127.0.0.1", port))
    except Exception:  # pylint: disable=broad-except
        sys.stderr.write("\nFailed to connect to copy tunnel, exiting.\n")
        sys.stderr.flush()
        return False

    count = 0
    stage = STAGE_NAME
    start = time.time()
    delta = start

    def show_progress():
        if count == 0:
            speed = "---.-- Kb"
        else:
            now = time.time()
            speed = format_size(count / (now - start))

        size = term_width - 54
        done = int(size * (count / filesize))

        sys.stdout.write("\r%35s [%s%s] %s/s " % (filename, "#" * done, "." * (size - done), speed))
        sys.stdout.flush()

    # Go!
    with open(filepath, "rb") as fp:
        while stage < STAGE_DONE:
            # pylint: disable=unused-variable
            rlist, wlist, xlist = select.select([sock], [], [], 0)

            if sock in rlist:
                data = sock.recv(CHUNK_SIZE)
                if data == CMD_ABORT:
                    # Aborting copy
                    break

                if data == b"OK":
                    stage += 1

                    sys.stdout.write("Ok.\n")
                    sys.stdout.flush()

            if time.time() > delta + TIMEOUT:
                break

            if stage == STAGE_DATA:
                data = fp.read(CHUNK_SIZE)
                if data:
                    sock.send(data)
                    filehash.update(data)
                    count += len(data)

                    if time.time() >= delta + 0.2:
                        show_progress()
                        delta = time.time()
                else:
                    show_progress()
                    delta = time.time()

                    stage += 1

            elif stage == STAGE_NAME:
                delta = time.time()

                sys.stdout.write("Sending name...")
                sys.stdout.flush()

                # Waiting for buffer to be flushed
                time.sleep(0.1)
                sock.send(to_bytes("%s:%s\n" % (CMD_NAME, filename)))

                stage += 1

            elif stage == STAGE_SIZE:
                delta = time.time()

                sys.stdout.write("Sending size...")
                sys.stdout.flush()

                # Waiting for buffer to be flushed
                time.sleep(0.1)
                sock.send(to_bytes("%s:%s\n" % (CMD_SIZE, filesize)))

                stage += 1

            elif stage == STAGE_HEAD:
                delta = time.time()

                sys.stdout.write("Sending data...\n\n")
                sys.stdout.flush()

                show_progress()
                stage += 1

            elif stage == STAGE_HASH:
                delta = time.time()

                sys.stdout.write("\nSending hash...")
                sys.stdout.flush()

                # Waiting for buffer to be flushed
                time.sleep(0.1)
                sock.send(to_bytes("%s:%s\n" % (CMD_HASH, filehash.hexdigest())))

                stage += 1

    if stage == STAGE_DONE:
        duration = time.time() - start
        speed = int(filesize / duration)

        sys.stdout.write(
            "\nTransfer completed of %s in %.02f second(s) at %s/s\n\n"
            % (format_size(filesize), duration, format_size(speed))
        )
    else:
        sys.stderr.write("\n\nAn error occurred during transfer.\n\n")

    try:
        sock.close()
    except Exception:  # pylint: disable=broad-except
        sys.stderr.write("\nFailed to close connection.\n")
        sys.stderr.flush()

    # Waiting remote for closing tunnel
    read_input(CMD_CLOSE, 2)

    return True


def format_size(size):
    """Format size specified as bytes"""
    if size > TB:
        size = size / TB
        unit = "TB"

    elif size > GB:
        size = size / GB
        unit = "GB"

    elif size > MB:
        size = size / MB
        unit = "MB"

    elif size > KB:
        size = size / KB
        unit = "KB"

    else:
        unit = "B"

    return "%6.02f %s" % (size, unit)


def main(args=None):
    """The script entrypoint"""
    parser = argparse.ArgumentParser(description="EeSSH copy client.")
    parser.add_argument(
        "filenames",
        metavar="file",
        nargs="+",
        help="A file to send to remote",
    )

    options = parser.parse_args(args)

    for filename in options.filenames:
        filepath = os.path.abspath(filename)
        copyfile(filepath)


if __name__ == "__main__":
    sys.exit(not main())
