# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Port Forwarding """

import hashlib
import logging
import os
import random
import select
import socket
import socketserver
import threading
import time

from paramiko.ssh_exception import SSHException

from eessh.eecopy import CHUNK_SIZE as EECOPY_CHUNK_SIZE

try:
    import Xlib.support.connect as xlib_connect
except ImportError:
    xlib_connect = None


log = logging.getLogger(__name__)


class PForwarder(threading.Thread, socketserver.ThreadingTCPServer):
    """Port forwarding daemon"""

    READ_SIZE = 1024

    def __init__(self, client, server_host, server_port, remote_host, remote_port, index=0):
        # Calling inherited
        threading.Thread.__init__(self, name=f"PFWD-{index:03d}")
        socketserver.ThreadingTCPServer.__init__(self, (server_host, server_port), None)

        # Non-blocking
        self.daemon = True
        self.daemon_threads = True
        self.allow_reuse_address = True

        self.index = index
        self.client = client

        self.remote_host = remote_host
        self.remote_port = remote_port

        self.__loop = True

    @property
    def server_host(self):
        """Get local listening address"""
        return self.server_address[0]

    @property
    def server_port(self):
        """Get local listening port"""
        return self.server_address[1]

    def close(self):
        """Close port forwarding"""
        self.__loop = False
        self.shutdown()
        return True

    stop = close

    def finish_request(self, request, client_address):
        """Processes the request"""
        transport = self.client.get_transport()

        try:
            chan = transport.open_channel(
                "direct-tcpip", (self.remote_host, self.remote_port), request.getpeername()
            )

        except Exception:  # pylint: disable=broad-except
            log.exception(
                "[PFWD] Incoming request to %s:%s failed.", self.remote_host, self.remote_port
            )
            return

        if chan is None:
            log.error(
                "[PFWD] Incoming request to %s:%s was rejected by the SSH server.",
                self.remote_host,
                self.remote_port,
            )
            return

        client_host, client_port = request.getpeername()
        server_host, server_port = chan.getpeername()

        log.info(
            "[PFWD] Tunnel opened from %s:%s forwarded from %s:%s to %s:%s",
            client_host,
            client_port,
            server_host,
            server_port,
            self.remote_host,
            self.remote_port,
        )

        while self.__loop:
            # pylint: disable=unused-variable
            rlist, wlist, xlist = select.select([request, chan], [], [], 0.1)

            if request in rlist:
                data = request.recv(self.READ_SIZE)
                if len(data) == 0:
                    break
                chan.send(data)

            if chan in rlist:
                data = chan.recv(self.READ_SIZE)
                if len(data) == 0:
                    break
                request.send(data)

        chan.close()
        request.close()

        log.info("[PFWD] Tunnel closed from %s:%s", client_host, client_port)

    def open(self):
        """Start port forwarding"""
        self.start()

        return True

    def run(self):
        """Run port forwarding daemon"""
        self.serve_forever()


class CForwardHandler(threading.Thread):
    """Copy file connection handler"""

    CMD_NAME = b"eessh-copy-name"
    CMD_SIZE = b"eessh-copy-size"
    CMD_HASH = b"eessh-copy-hash"
    CMD_DONE = b"eessh-copy-done"

    STAGE_NAME = 0
    STAGE_SIZE = 1
    STAGE_DATA = 2
    STAGE_HASH = 3
    STAGE_DONE = 4

    def __init__(self, channel, dirpath, timeout=10):
        # Calling inherited
        super().__init__()

        # Non-blocking
        self.daemon = True

        self.channel = channel
        self.dirpath = dirpath
        self.timeout = timeout

        self.__loop = True

    def run(self):
        """Run the copy file forwarding handler"""
        filehash = None
        filename = None
        filesize = None
        recvhash = hashlib.sha256()

        count = 0
        file_ = None
        stage = self.STAGE_NAME
        start = time.time()
        delta = start

        while self.__loop and stage < self.STAGE_DONE:
            # pylint: disable=unused-variable
            rlist, wlist, xlist = select.select([self.channel], [], [], 0.1)

            if self.channel not in rlist:
                if time.time() > delta + self.timeout:
                    log.error("[CFWD] Timeout.")
                    break
                continue

            data = self.channel.recv(EECOPY_CHUNK_SIZE)
            if len(data) == 0:
                if stage != self.STAGE_DONE:
                    log.error("[CFWD] An error occurred during transfer.")
                break

            delta = time.time()

            if stage == self.STAGE_DATA:
                count += len(data)
                file_.write(data)
                recvhash.update(data)

                if count == filesize:
                    self.channel.send(b"OK")

                    stage += 1

                    recvhash = recvhash.hexdigest()

                    log.info("[CFWD] recv=%s", recvhash)

            elif stage == self.STAGE_NAME:
                if data.startswith(self.CMD_NAME + b":"):
                    self.channel.send(b"OK")

                    stage += 1

                    filename = data.decode("utf-8").strip().split(":", 1)[1]
                    filepath = os.path.join(self.dirpath, filename)
                    # pylint: disable=consider-using-with
                    file_ = open(filepath, "wb")

                    log.info("[CFWD] name=%s...", filename)

            elif stage == self.STAGE_SIZE:
                if data.startswith(self.CMD_SIZE + b":"):
                    self.channel.send(b"OK")

                    stage += 1

                    filesize = int(data.decode("utf-8").strip().split(":", 1)[1])

                    log.info("[CFWD] size=%s", filesize)

            elif stage == self.STAGE_HASH:
                if data.startswith(self.CMD_HASH + b":"):
                    self.channel.send(b"OK")

                    stage += 1

                    filehash = data.decode("utf-8").strip().split(":", 1)[1]
                    if filehash == recvhash:
                        log.info("[CFWD] sent=%s - OK.", filehash)
                    else:
                        log.error("[CFWD] sent=%s - sum mismatch, transfer failed!", filehash)

        if file_:
            file_.close()

    def stop(self):
        """Stop the copy file forwarding handler"""
        self.__loop = False


class RForwardHandler(threading.Thread):
    """Reverse port forwarding connection handler"""

    READ_SIZE = 1024

    def __init__(self, channel, remote_host, remote_port):
        # Calling inherited
        super().__init__()

        # Non-blocking
        self.daemon = True

        self.channel = channel
        self.remote_host = remote_host
        self.remote_port = remote_port

        self.__loop = True

    def run(self):
        """Run the reverse port forwarding handler"""
        sock = socket.socket()

        try:
            sock.connect((self.remote_host, self.remote_port))

        except Exception:  # pylint: disable=broad-except
            log.exception(
                "[RFWD] Failed forwarding request to %s:%s", self.remote_host, self.remote_port
            )
            self.channel.close()
            return

        client_host, client_port = self.channel.origin_addr
        server_host, server_port = self.channel.getpeername()

        log.info(
            "[RFWD] Tunnel opened from %s:%s forwarded from %s:%s to %s:%s",
            client_host,
            client_port,
            server_host,
            server_port,
            self.remote_host,
            self.remote_port,
        )

        while self.__loop:
            # pylint: disable=unused-variable
            rlist, wlist, xlist = select.select([sock, self.channel], [], [], 0.1)

            if sock in rlist:
                data = sock.recv(self.READ_SIZE)
                if len(data) == 0:
                    break
                self.channel.send(data)

            if self.channel in rlist:
                data = self.channel.recv(self.READ_SIZE)
                if len(data) == 0:
                    break
                sock.send(data)

        self.channel.close()
        sock.close()

        log.info("[RFWD] Tunnel closed from %s:%s", client_host, client_port)

    def stop(self):
        """Stop the reverse port forwarding handler"""
        self.__loop = False


class RForwarderMixin:
    """Reverse port forwarding mixin"""

    key = "RFWD"
    name = "Unset name"

    __forwarders = {}

    def __init__(self, client, server_host, server_port, index=0):
        self.index = index
        self.client = client
        self.handlers = []

        self.server_host = server_host
        self.server_port = server_port

    @classmethod
    def handler(cls, channel, origin_addr, server_addr):
        """Build and start a port forwarding handler"""
        client_host, client_port = origin_addr
        server_host, server_port = server_addr

        log.info(
            "[RFWD] channel=%s. client=%s:%s, server=%s:%s",
            channel.get_id(),
            client_host,
            client_port,
            server_host,
            server_port,
        )

        self = cls.__forwarders.get(server_addr)
        if self is None:
            # Oops
            return

        handler = self.get_handler(channel, origin_addr, server_addr)
        handler.start()

        self.handlers.append(handler)

    def close(self):
        """Close all"""
        for handler in self.handlers:
            handler.stop()

        # De-register forwarder
        del self.__forwarders[(self.server_host, self.server_port)]

        try:
            transport = self.client.get_transport()
            transport.cancel_port_forward(self.server_host, self.server_port)

            if self.__forwarders:
                # Handler is removed when calling .cancel_port_forward() method
                # Set it again for other forwarders
                transport._tcp_handler = self.handler  # pylint: disable=protected-access
        except SSHException:
            log.exception("[%s] Failed to close %s", self.key, self.name)
            return False

        log.info("[%s] Closed %s", self.key, self.name)
        return True

    def get_handler(self, channel, origin_addr, server_addr):
        """Get the port forwarding handler"""
        raise NotImplementedError()

    def open(self):
        """Start port forwarding"""
        try:
            transport = self.client.get_transport()
            transport.request_port_forward(self.server_host, self.server_port, handler=self.handler)
        except SSHException:
            log.exception("[%s] Failed to open %s", self.key, self.name)
            return False

        log.info("[%s] Opening %s", self.key, self.name)

        # Register forwarder
        self.__forwarders[(self.server_host, self.server_port)] = self

        return True


class RForwarder(RForwarderMixin):
    """Reverse port forwarding object"""

    def __init__(self, client, server_host, server_port, remote_host, remote_port, index=0):
        # Calling inherited
        super().__init__(client, server_host, server_port, index=index)

        self.remote_host = remote_host
        self.remote_port = remote_port

    @property
    def name(self):
        """Retrieve the name of the port forwarding object"""
        return f"port forwarding tunnel from {self.server_host}:{self.server_port} to {self.remote_host}:{self.remote_port}"

    def get_handler(self, channel, origin_addr, server_addr):
        """Get the reverse port forwarding handler"""
        return RForwardHandler(channel, self.remote_host, self.remote_port)


class CForwarder(RForwarderMixin):
    """Copy file object"""

    CMD_START = b"eessh-copy-start"
    CMD_CLOSE = b"eessh-copy-close"
    CMD_READY = b"eessh-copy-ready"

    DATA_PORTS = (23000, 24000)

    key = "CFWD"

    def __init__(self, client, server_port=None, dirpath=None):
        if server_port is None:
            server_port = random.randint(*self.DATA_PORTS)  # nosec

        # Calling inherited
        super().__init__(client, "127.0.0.1", server_port)

        self.dirpath = dirpath if dirpath is not None else os.getcwd()

    @property
    def name(self):
        """Retrieve the name of the copy file object"""
        return f"copy tunnel from {self.server_host}:{self.server_port} to {self.dirpath}"

    def get_handler(self, channel, origin_addr, server_addr):
        """Get the copy file handler"""
        return CForwardHandler(channel, self.dirpath)


class XForwardHandler(threading.Thread):
    """X11 forwarding connection handler"""

    def __init__(self, chan, disp):
        # Calling inherited
        super().__init__()

        # Non-blocking
        self.daemon = True

        self.chan = chan
        self.disp = disp
        self.sock = None

        self.__loop = True

    def run(self):
        """Run the X11 forwarding connection handler"""
        try:
            self.sock = xlib_connect.get_socket(*self.disp[:4])

        except Exception:  # pylint: disable=broad-except
            log.exception("[XFWD] Forwarding request failed")
            return

        log.info("[XFWD] Tunnel opened")

        while self.__loop:
            # pylint: disable=unused-variable
            rlist, wlist, xlist = select.select([self.sock, self.chan], [], [], 0.1)

            if self.sock in rlist:
                data = self.sock.recv(4096)
                if len(data) == 0:
                    break
                self.chan.sendall(data)

            if self.chan in rlist:
                data = self.chan.recv(4096)
                if len(data) == 0:
                    break
                self.sock.sendall(data)

        self.chan.close()
        self.sock.close()

        log.info("[XFWD] Tunnel closed")

    def stop(self):
        """Stop the X11 forwarding connection handler"""
        self.__loop = False


class XForwarder(threading.Thread):
    """X11 forwarder object"""

    def __init__(self, channel, display):
        # Calling inherited
        threading.Thread.__init__(self, name="XFWD")

        # Non-blocking
        self.daemon = True

        self.__loop = True
        self.__error = None
        self.__ready = False

        self.channel = channel
        self.display = display

    @property
    def ready(self):
        """Check if ready"""
        return self.__ready

    def close(self):
        """Close forwarder"""
        self.__loop = False

    stop = close

    def open(self):
        """Open forwarder"""
        self.start()

        # Wait for tunnel to be ready
        while not self.__ready and self.__error is None:
            time.sleep(0.1)

        return self.__error is None

    def run(self):
        """Start forwarding"""
        if xlib_connect is None:
            self.__ready = True

            log.error("[XFWD] Please install Xlib module in order to use X11 forwwarding, exiting.")
            return

        display = xlib_connect.get_display(self.display)

        handlers = []

        # pylint: disable=unused-argument,dangerous-default-value
        def handler(chan, origin_addr, display=display, handlers=handlers):
            handler = XForwardHandler(chan, display)
            handler.start()

            handlers.append(handler)

        try:
            self.channel.request_x11(handler=handler)

        except SSHException as err:
            log.exception("[XFWD] Failed to open X11 forwarding tunnel")
            self.__error = err
            return

        log.info("[XFWD] Opening X11 forwarding tunnel")

        self.__ready = True
        while self.__loop:
            # Don't eat CPU
            time.sleep(0.1)

        for handler in handlers:
            handler.stop()
