# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Legacy connector using XML-RPC """

import logging
import threading
import xmlrpc.client
from base64 import b64encode

from eessh.cache import xdgcache as region

log = logging.getLogger(__name__)


# Errors
ERROR_UNKNOWN = 10000
ERROR_ACCES_DENIED = 10001
ERROR_TRACEROUTE_NO_SERVER = 10010
ERROR_TRACEROUTE_NO_ROUTE = 10011

marshaller = xmlrpc.client.Marshaller(allow_none=True)


def dump_struct(data):
    """Simple function to convert a dictionary to a string"""

    out = []

    marshaller.dump_struct(data, out.append)

    return "".join(out)


class LegacyConnectorError(xmlrpc.client.Error):
    """
    Legacy XML-RPC Connector error
    """

    def __init__(self):
        xmlrpc.client.Error.__init__(self)
        self.errmsg = "Unknown error"
        self.errcode = -1

    def __repr__(self):
        return f"<LegacyConnectorError: {self.errmsg} {self.errcode}>"


class LegacyConnector:
    """
    Legacy XML-RPC Connector to Eemyle server
    """

    def __init__(
        self, url, username=None, password=None, debug=False, **kwargs
    ):  # pylint: disable=unused-argument
        def basic_auth(username, password):
            token = b64encode(f"{username}:{password}".encode()).decode("ascii")
            return f"Basic {token}"

        # Build server
        self.__error = None
        self.__server = xmlrpc.client.Server(
            url,
            headers=[("Authorization", basic_auth(username, password))],
            verbose=debug,
        )

        self.__lock = threading.Lock()

    def __enter__(self):
        """Acquire lock"""
        self.__lock.acquire()
        return self

    def __exit__(self, type_, value, traceback):
        """Release lock"""
        self.__lock.release()

    @property
    def error(self):
        """Retrieve last error"""
        return self.__error

    @region.cache_on_arguments_if_not_refresh()
    def get_customers(self):
        """Retrieve customers info"""
        result = self.query("eemyle.get_customers_having_server")
        if result is None:
            return None

        customers = []
        for data in result:
            if not data["has_server"]:
                continue

            customers.append(
                {
                    "customer_uid": data["eeid"],
                    "name": data["name"],
                    "city": data["city"],
                    "address": data["street"],
                    "phone": data["phone"],
                    "has_server": data["has_server"],
                }
            )

        return customers

    @region.cache_on_arguments_if_not_refresh(leave_refresh_kwarg=True)
    def _get_server_id(self, customer_uid, hostname, refresh=False):
        """Retrieve server ID from customer UID and server hostname"""
        # pylint: disable=unexpected-keyword-arg
        servers = self.get_servers(customer_uid, refresh=refresh)
        if servers is None:
            log.warning(
                "_get_server_id(%s, %s): failed to list customer servers", customer_uid, hostname
            )
            return None

        for server in servers:
            if server["hostname"] == hostname:
                return str(server["id"])

        log.warning(
            "_get_server_id(%s, %s): server not found in list of customer servers",
            customer_uid,
            hostname,
        )
        return None

    def get_data(self, customer_uid=None, hostname=None, server_id=None, refresh=False):
        """Retrieve server extended info"""
        if server_id is None:
            server_id = self._get_server_id(customer_uid, hostname, refresh=refresh)
            if server_id is None:
                return None

        if refresh:
            # Invalidate server caches
            self._invalidate_server_cache(server_id)

        # Get server
        data = self.get_server(server_id)
        if data is None:
            return None

        data.update(
            {
                "hops": self.get_hops(server_id),
                "ipmi": self.get_ipmi(server_id),
            }
        )

        return data

    def _invalidate_server_cache(self, server_id):
        """Invalidate server data cached"""
        log.debug("Invalidate server #%s cache", server_id)
        self.get_server.invalidate(self, server_id)  # pylint: disable=no-member
        self.get_hops.invalidate(self, server_id)  # pylint: disable=no-member
        self.get_ipmi.invalidate(self, server_id)  # pylint: disable=no-member

    @region.cache_on_arguments()
    def get_hops(self, server_id):
        """Retrieve server connection hops"""
        result = self.query("eemyle.get_server_traceroute_by_server_id", str(server_id))
        if not isinstance(result, list):
            return None

        # Prepare data
        hops = []
        for data in result:
            hop = {
                "hostname": data.get("server_name"),
                "host": data.get("ip"),
                "port": int(data.get("ssh_port", "22")),
                "username": data.get("admin_name"),
                "password": data.get("admin_pass") or None,
            }

            if "sudo" in data["root_pass"] or data["sudo_root"] == "t":
                hop["su_method"] = "sudo"
                hop["su_username"] = data["root_name"]
                hop["su_password"] = data["admin_pass"] or None

            elif data["root_name"] == "enable":
                hop["su_method"] = "enable"
                hop["su_password"] = data["root_pass"] or None

            else:
                hop["su_method"] = "su"
                hop["su_username"] = data["root_name"]
                hop["su_password"] = data["root_pass"] or None

            hops.append(hop)

        return hops

    @region.cache_on_arguments()
    def get_ipmi(self, server_id):
        """Retrieve server IPMI info"""
        result = self.query("eemyle.get_server_ipmi_infos_by_server_id", str(server_id))
        if result is None:
            return None

        ipmi = {
            "host": result["host"] or None,
            "port": 623,
            "username": result["user"] or None,
            "password": result["password"] or None,
        }

        ipmi["enabled"] = (
            ipmi["host"] is not None
            and ipmi["username"] is not None
            and ipmi["password"] is not None
        )

        return ipmi

    @region.cache_on_arguments()
    def get_server(self, server_id):
        """Retrieve server info"""
        assert server_id and str(server_id) not in ("0", "-1"), f'Invalid server ID "{server_id}"'

        result = self.query("eemyle.get_server_by_server_id", str(server_id))
        if result is None:
            return None
        log.debug("get_server(%s): API data=%s", server_id, result)

        data = {
            "id": int(result["id_server"]),
            "customer_uid": result["id_customer"],
            "hostname": result["server_name"].split()[0],
            "description": result["usage"],
            "comments": result["comments"],
            "security": result["security_check"] == "t",
            "interfaces": [
                {"id": _id, "name": name} for _id, name in result["connection_interface_data"]
            ],
            "physical_host": None,
            "users": [
                {
                    "username": user["name"],
                    "password": user["pass"] or None,
                }
                for user in result["users"]
            ],
        }
        if result["id_server_physical_host"] and result["id_server_physical_host"] not in (
            "0",
            "-1",
        ):
            data["physical_host"] = self.get_server(result["id_server_physical_host"])

        if result["id_connection_interface"]:
            for interface in data["interfaces"]:
                if interface["id"] == result["id_connection_interface"]:
                    data["connection_interface"] = interface
        return data

    def update_data(self, changes, customer_uid=None, hostname=None, server_id=None):
        """Update server info"""
        if server_id is None:
            server_id = self._get_server_id(customer_uid, hostname)
            if server_id is None:
                return False

        mapped_changes = {}
        simple_map = {"hostname": "server_name", "description": "usage", "comments": "comments"}
        for key, value in changes.items():
            if key not in simple_map:
                log.error(
                    'update_data(%s): change on "%s" not supported/invalid key', server_id, key
                )
                return False
            mapped_changes[simple_map[key]] = value

        log.debug("Changes: %s", mapped_changes)
        result = self.query("eemyle.update_server", str(server_id), mapped_changes)
        log.debug("update_data(%s): result = %s", server_id, result)
        if result != 1:
            return False

        # Invalidate cache data of server
        self._invalidate_server_cache(server_id)

        return True

    @region.cache_on_arguments_if_not_refresh()
    def get_servers(self, customer_uid):
        """Retrieve customer servers"""
        if not customer_uid:
            return None

        result = self.query("eemyle.get_servers_by_customer_id", customer_uid)
        if result is None:
            return None

        servers = []
        for data in result:
            if data["dead"] == "t":
                continue

            servers.append(
                {
                    "id": int(data["id_server"]),
                    "hostname": data["server_name"].split()[0],
                    "hoster": None,
                    "os_name": data["os_name"],
                    "description": data["usage"],
                    "security": data["security_check"] == "t",
                    "usage": ", ".join(sorted(data["usages"].values()) if data["usages"] else []),
                    "usernames": (sorted(data["users"].values()) if data["users"] else []),
                }
            )

        return servers

    @region.cache_on_arguments_if_not_refresh(leave_refresh_kwarg=True)
    def get_tickets(self, customer_uid=None, hostname=None, server_id=None, refresh=False):
        """Retrieve server tickets"""
        if server_id is None:
            server_id = self._get_server_id(customer_uid, hostname, refresh=refresh)
            if server_id is None:
                return None

        result = self.query("eemyle.get_tickets_by_server_id", str(server_id))
        if result is None:
            return None

        tickets = []
        for data in result:
            tickets.append(data)

        return tickets

    def query(self, function_name, *args):
        """Sends a request to the server"""

        log.debug("Query function %s(%s)", function_name, ", ".join([str(arg) for arg in args]))

        self.__error = None

        # Encoding the arguments
        enc_args = []
        for arg in args:
            if isinstance(arg, dict):
                enc_arg = dump_struct(arg)
            elif isinstance(arg, int):
                enc_arg = str(arg)
            else:
                try:
                    enc_arg = xmlrpc.client.Binary(bytes(arg, "utf-8"))
                except Exception:  # pylint: disable=broad-except
                    log.exception(
                        "Failed to encode argument %s of type %s for function %s(%s)",
                        arg,
                        type(arg),
                        function_name,
                        ",".join([str(i) for i in args]),
                    )
                    enc_arg = str(arg)

            enc_args.append(enc_arg)

        # Calling the function
        try:
            function = getattr(self.__server, function_name)
            result = function(*enc_args)

        except xmlrpc.client.Error as err:
            # Xmlrpclib error
            log.exception(
                "Failed to call xmlrpc function %s(%s)",
                function_name,
                ",".join([str(i) for i in args]),
            )

            self.__error = err
            result = None

        except Exception:  # pylint: disable=broad-except
            log.exception(
                "Unexpected exception occurred calling xmlrpc function %s(%s)",
                function_name,
                ",".join([str(i) for i in args]),
            )

            # Unknown error
            self.__error = LegacyConnectorError()
            result = None

        return result
