# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Shell """

import logging
import os
import os.path
import shlex
import tempfile
from functools import cmp_to_key

from cmd2 import Cmd

from eessh.config import BooleanOption, IntegerOption, StringOption, config
from eessh.utils import (
    attr_repr,
    attr_sort,
    get_host_port,
    pretty_format_pfwd,
    pretty_format_server_data,
    pretty_format_server_passwords,
    pretty_format_server_tickets,
    resolv_local_paths,
)

log = logging.getLogger(__name__)

COMMAND_PROMPT = "(\x1b[34meessh\x1b[39m) {username}@{hostname}:{path} > "

# Register config section & options
section = config.add_section("shell", comment="EeSSH shell", order=4)
section.add_option(
    BooleanOption, "debug", default=False, comment="Enable debug mode on starting shell"
)
section.add_option(
    StringOption,
    "initial_remote_directory",
    comment="Initial remote directory of the SFTP connection on the remote host",
)
section.add_option(
    StringOption,
    "history_path",
    comment="Shell history file path (default: history file on same directory of the configuration file)",
)
section.add_option(
    IntegerOption,
    "history_length",
    default=1000,
    comment="Max number of executed commands to keep in history file",
)


class EeCMDShell(Cmd):
    """EeSSH command shell"""

    allow_cli_args = False

    def __init__(self, client):
        self.client = client

        if config.get("shell", "debug"):
            log.debug("Enable debug mode (from config)")
            self.debug = True

        # Make sure history directory exists
        self.history_path = os.path.expanduser(
            config.get("shell", "history_path") or os.path.join(config.config_dir, "history")
        )
        log.debug("Use %s history file", self.history_path)
        history_dirpath = os.path.dirname(self.history_path)
        if not os.path.isdir(history_dirpath):
            try:
                os.makedirs(history_dirpath)
            except OSError:
                pass

        # Calling inherited
        Cmd.__init__(
            self,
            persistent_history_file=self.history_path,
            persistent_history_length=config.get("shell", "history_length"),
        )

        # Handle initial remote path
        self.path = config.get("shell", "initial_remote_directory")
        if self.path and not self.client.do_cd(self.path):
            # On fail, just retrieve current remote path
            log.warning("Failed to change directory on remote host (%s)", self.path)
            self.path = self.client.do_pwd()
        else:
            self.path = self.client.do_pwd()

        self._update_prompt()

    def _update_prompt(self):
        """Update shell prompt"""
        self.prompt = COMMAND_PROMPT.format(
            username=self.client.username,
            hostname=self.client.hostname,
            path=self.path if self.path else "?",
        )

    def _complete_remote(self, text, line, begidx, endidx, dir_only=False):
        """Handle argument completion for remote file/directory path"""
        log.debug(
            '[COMPLETE] type=remote, text=%s, line="%s", begin/end idx=%d/%d, dir only=%s',
            text,
            line,
            begidx,
            endidx,
            dir_only,
        )

        # Extract directory from path
        dirpath = os.path.dirname(text) or "."
        basename = os.path.basename(text)

        log.debug(
            '[COMPLETE] type=remote, text=%s, dir path="%s", basename="%s"', dirpath, text, basename
        )

        # List directory
        data = self.client.do_ls(dirpath)
        log.debug("[COMPLETE] type=remote, content data=%s", data)
        if data is None:
            log.debug('[COMPLETE] type=remote, path "%s" does not exists', dirpath)
            return []

        # Filter result
        result = []
        log.debug('[COMPLETE] type=remote, path "%s", iter on content', dirpath)
        for attr in data:
            filename = attr["filename"]
            log.debug("[COMPLETE] type=remote, filename=%s, attrs=%s", filename, attr)
            if basename and not filename.startswith(basename):
                continue

            fullpath = os.path.join(dirpath, filename)
            is_dir = attr["type"] == "d"
            if is_dir:
                fullpath += os.sep
            elif dir_only:
                continue

            result.append(fullpath)

        if len(result) == 1 and is_dir:
            self.allow_appended_space = False
            self.allow_closing_quote = False
        else:
            self.allow_appended_space = True
            self.allow_closing_quote = True

        log.debug("[COMPLETE] type=remote, result=%s", result)
        return result

    def _complete_one_remote(self, text, line, begidx, endidx):
        """Handle arguments completion for command accepting only one remote file path"""
        return self.index_based_complete(
            text, line, begidx, endidx, index_dict={1: self._complete_remote}
        )

    def _complete_multiple_remote(self, text, line, begidx, endidx):
        """Handle arguments completion for command accepting multiple remote file path"""
        return self.index_based_complete(
            text, line, begidx, endidx, index_dict={}, all_else=self._complete_remote
        )

    def _complete_one_remote_directory(self, text, line, begidx, endidx):
        """Handle arguments completion for command accepting only one remote directory path"""
        return self.index_based_complete(
            text,
            line,
            begidx,
            endidx,
            index_dict={1: lambda *args: self._complete_remote(*args, dir_only=True)},
        )

    def _complete_remote_or_local(self, text, line, begidx, endidx):
        """Handle argument completion for remote or local file path"""
        local_result = self.path_complete(text, line, begidx, endidx) or []
        remote_result = self._complete_remote(text, line, begidx, endidx) or []
        return local_result + remote_result

    complete_cat = _complete_multiple_remote
    complete_ls = _complete_multiple_remote
    complete_cd = _complete_one_remote_directory

    def complete_get(self, text, line, begidx, endidx):
        """Handle get command arguments completion"""
        log.debug(
            '[COMPLETE] get, text="%s", line="%s", begin/end idx=%d / %d',
            text,
            line,
            begidx,
            endidx,
        )
        return self.index_based_complete(
            text,
            line,
            begidx,
            endidx,
            index_dict={
                1: self._complete_remote,
            },
            all_else=self._complete_remote_or_local,
        )

    def complete_put(self, text, line, begidx, endidx):
        """Handle put command arguments completion"""
        log.debug(
            '[COMPLETE] put, text="%s", line="%s", begin/end idx=%d / %d',
            text,
            line,
            begidx,
            endidx,
        )
        return self.index_based_complete(
            text,
            line,
            begidx,
            endidx,
            index_dict={
                1: self.path_complete,
            },
            all_else=self._complete_remote_or_local,
        )

    def complete_pfwd(
        self, text, line, begidx, endidx, opened_pfwd=None
    ):  # pylint: disable=unused-argument
        """Handle pfwd command arguments completion"""

        args = shlex.split(line)
        match_against = []

        if len(args) == 1 or (len(args) == 2 and text):
            match_against = ["close", "list", "open"]

        elif (len(args) == 2 and args[1] == "close") or (len(args) == 3 and text):
            match_against = ["all"] + [
                str(pfwd[0])
                for pfwd in (opened_pfwd if opened_pfwd is not None else self.client.pfwds)
            ]

        return self.basic_complete(text, line, begidx, endidx, match_against=match_against)

    def complete_rfwd(self, *args):  # pylint: disable=unused-argument
        """Handle rfwd command arguments completion"""

        return self.complete_pfwd(*args, opened_pfwd=self.client.rfwds)

    def do_su(self, arg):  # pylint: disable=unused-argument
        """Become super user"""

        # Command will be performed by client itself
        self.lastcmd = "su"
        return True

    @staticmethod
    def do_exit(arg):  # pylint: disable=unused-argument
        """Exit from command mode"""

        return True

    def do_cat(self, arg):
        """
        Display content of remote file(s) (using SFTP)

        Usage:
            cat <remote file path> [<other remote file path>]
        """

        paths = shlex.split(arg)

        if not paths:
            self.perror("Missing argument.")
            return

        for path in paths:
            content = self.client.do_cat(path)
            if content is None:
                self.perror("Failed to open file.")
                return

            content = content.decode("utf8", errors="ignore")

            if len(paths) > 1:
                self.poutput(f"{path}:")

            self.ppaged(content)

            if len(paths) > 1:
                self.poutput("\n")

    def do_cd(self, arg):
        """cd <path> - change current directory on remote (SFTP)"""

        args = shlex.split(arg)

        path = args[0] if args else None
        if path == "~":
            path = None

        if not self.client.do_cd(path):
            self.perror("Failed to change directory.")
            return

        self.path = "~" if path is None else os.path.normpath(os.path.join(self.path, path))

        # Update prompt
        self._update_prompt()

    def do_get(self, arg):
        """
        Retrieve file(s) from remote (SFTP)

        Usage:
            get [remote path] [other remote path] [local path]

        local path: optional, default: current local directory
        """

        args = shlex.split(arg)

        if not args:
            self.perror("Missing argument.")
            return

        remotepaths = args[:-1] if len(args) > 1 else args
        localpath = args[-1] if len(args) > 1 else None

        for remotepath in remotepaths:
            if not self.client.do_get(remotepath, localpath):
                self.perror(f'Failed to get "{remotepath}" file.')

    def do_ls(self, arg):
        """
        List files on remote (SFTP)

        Usage:
            ls [remote path]

        remote path: Remote path to list (optional, default: current remote directory)
        """

        paths = shlex.split(arg if arg else ".")
        paths = [path if path else "." for path in paths]

        result = []
        for path in paths:
            data = self.client.do_ls(path)
            if data is None:
                self.perror(f'Failed to list directory "{path}".')
                continue

            if len(paths) > 1:
                result.append(f"{path}:")

            for attr in sorted(data, key=cmp_to_key(attr_sort)):
                result.append(attr_repr(attr))

            if len(paths) > 1:
                result.append("\n")
        if result:
            self.ppaged("\n".join(result))

    def do_put(self, arg):
        """
        Transfer file(s) to remote (SFTP)

        Usage:
            put <local_file> [<other_local_file>] <remote_path>

        remote_path: optional, default: current remote directory
        """

        args = shlex.split(arg)

        if not args:
            self.perror("Missing argument.")
            return

        localpaths = args[:-1] if len(args) > 1 else args
        remotepath = args[-1] if len(args) > 1 else None

        for localpath in localpaths:
            for path in resolv_local_paths(localpath):
                if not path.exists():
                    self.perror(f"{localpath}: not found")
                    continue
                if path.is_dir():
                    self.perror("%s: transferring directory is not supported")
                    continue
                if not self.client.do_put(str(path), remotepath):
                    self.perror(f'Failed to put "{localpath}" file in "{remotepath}".')

    def do_pwd(self, arg):  # pylint: disable=unused-argument
        """Display current directory on remote (SFTP)"""

        path = self.client.do_pwd()

        self.poutput(f"{path}\n\n")

    def do_pfwd(self, arg):
        """
        Open or close a port forwarding tunnel with remote

        Usage:
            pfwd [list]
            pfwd open [remote_host:]remote_port [local_host:][local_port]
            pfwd close index

        remote_host: optional remote host to connect to, default is to connect to localhost
        remote_port: remote port to connect to
        local_host: optional local host to listen on, default is to listen on localhost
        local_port: optional local port to listen on, default is same as remote port

        Without arguments, `pfwd' list current opened tunnel(s).
        """

        args = shlex.split(arg)

        if not args or args[0] == "list":
            pfwds = self.client.pfwds
            if not pfwds:
                self.poutput("No tunnel.")
                return

            self.ppaged(pretty_format_pfwd(pfwds))

        elif args[0] == "close":
            if len(args) < 2:
                self.poutput("Missing argument.")
            else:
                self.client.close_pfwd(args[1])

        elif args[0] == "open":
            if len(args) < 2:
                self.perror("Missing argument.")
            else:
                try:
                    remote_host, remote_port = get_host_port(args[1])
                    local_host, local_port = (
                        get_host_port(args[2]) if len(args) > 2 else ("127.0.0.1", remote_port)
                    )

                except ValueError:
                    self.perror("Invalid argument.")

                else:
                    self.client.open_pfwd(local_host, local_port, remote_host, remote_port)
        else:
            self.perror("Invalid argument.")

    def do_rfwd(self, arg):
        """
        Open or close a reverse port forwarding tunnel with remote

        Usage:
            rfwd [list]
            rfwd open [local_host:]local_port [remote_host:][remote_port]
            rfwd close index

        local_host: optional local host to connect to, default is to connect to localhost
        local_port: local port to connect to
        remote_host: optional remote host to listen on, default is to listen on localhost
        remote_port: optional remote port to listen on, default is same as local port

        Without arguments, `rfwd' list current opened tunnel(s).
        """

        args = shlex.split(arg)

        if not args or args[0] == "list":
            rfwds = self.client.rfwds
            if not rfwds:
                self.poutput("No tunnel.")
                return

            self.ppaged(pretty_format_pfwd(rfwds, reverse=True))

        elif args[0] == "close":
            if len(args) < 2:
                self.perror("Missing argument.")
            else:
                self.client.close_rfwd(args[1])

        elif args[0] == "open":
            if len(args) < 2:
                self.perror("Missing argument.")
            else:
                try:
                    local_host, local_port = get_host_port(args[1])
                    remote_host, remote_port = (
                        get_host_port(args[2]) if len(args) > 2 else ("127.0.0.1", local_port)
                    )

                except ValueError:
                    self.perror("Invalid argument.")

                else:
                    self.client.open_rfwd(remote_host, remote_port, local_host, local_port)
        else:
            self.perror("Invalid argument.")

    def do_show(self, raw_args):
        """
        Show server information

        Usage:
            show [info|passwords|tickets|comments] [--refresh]
        """

        info = None
        refresh = False
        for arg in shlex.split(raw_args):
            if arg == "--refresh":
                refresh = True
            elif arg not in ("info", "tickets", "passwords", "comments"):
                self.perror(
                    f'Invalid information "{arg}". Must be one of following: info, passwords, tickets or comments'
                )
                return
            elif info:
                self.perror("You can request only one information at a time.")
                return
            else:
                info = arg

        # Handle default info
        if not info:
            info = "info"

        content = None
        if info == "tickets":
            tickets = self.client.get_server_tickets(refresh=refresh)
            if not tickets:
                self.perror("Failed to retrieve tickets about current server")
                return

            content = pretty_format_server_tickets(tickets)
        else:
            data = self.client.get_server_data(refresh=refresh)
            log.debug("show(%s): data=%s", info, data)
            if not data:
                self.perror("Failed to retrieve data about current server")
                return

            if info == "info":
                content = pretty_format_server_data(data)
            elif info == "passwords":
                content = pretty_format_server_passwords(data)
            elif info == "comments":
                content = (
                    "Comments:\n" + "---------\n" + data.get("comments", "No comment.") + "\n\n"
                )

        self.ppaged(content)

    def complete_show(self, text, line, begidx, endidx):
        """Handle show command arguments completion"""
        return self.index_based_complete(
            text,
            line,
            begidx,
            endidx,
            index_dict={},
            all_else=("info", "tickets", "passwords", "comments", "--refresh"),
        )

    def do_edit(self, args):
        """
        Edit server information

        Usage:
            edit [comments]
        """

        args = shlex.split(args)
        if len(args) != 1:
            self.perror("Invalid number of argument")
            return

        info = args[0]
        if info != "comments":
            self.perror(f'Invalid information "{info}". Only "comments" is currently supported.')
            return

        # Retrieve current up to date data
        data = self.client.get_server_data(refresh=True)
        log.debug("edit(%s): data=%s", info, data)
        if not data:
            self.perror("Failed to retrieve data about current server")
            return

        path = None

        def drop_tmpfile():
            if path and os.path.exists(path):
                try:
                    os.remove(path)
                except Exception:  # pylint: disable=broad-except
                    log.exception(
                        'An exception occurred removing temporary file "%s". Consider dropping it manually.',
                        path,
                    )

        # Store comment in temporary file
        try:
            path = tempfile.mkstemp()
            log.debug("edit(%s): temporary file = %s", info, path)
            with open(path, "w", encoding="utf8") as file_desc:
                file_desc.write(data["comments"])
        except Exception:  # pylint: disable=broad-except
            log.exception(
                "edit(%s): An exception occurred storing current comments in temporary file", info
            )
            self.perror("Failed to store current comments in temporary file")
            drop_tmpfile()
            return

        # Edit temporary file with editor
        try:
            super().do_edit(path)
        except Exception:  # pylint: disable=broad-except
            log.exception("An exception occurred editing temporary file")
            self.perror("Error editing temporary file")
            drop_tmpfile()
            return

        # Double-switch buffer to clear screen after running external editor
        self.client.switch_buffer()
        self.client.switch_buffer()

        # Retrieve new comments
        with open(path, encoding="utf8") as file_desc:
            new_comments = file_desc.read()

        log.debug(
            "edit(%s): new comments):\n\n  %s",
            data["hostname"],
            "\n  ".join(new_comments.split("\n")),
        )

        if new_comments == data["comments"]:
            self.poutput("No change")
            drop_tmpfile()
            return

        result = self.client.update_data({"comments": new_comments})
        log.debug("edit(%s): update data result = %s", info, result)
        if result:
            self.poutput("Your changes have been saved.")
            drop_tmpfile()
        else:
            self.perror(
                f"Failed to save your changes. Temporary file with your change keep: {path}"
            )

    def complete_edit(self, text, line, begidx, endidx):
        """Handle edit command arguments completion"""
        return self.index_based_complete(
            text,
            line,
            begidx,
            endidx,
            index_dict={
                1: lambda *args, **kwargs: self.basic_complete(
                    *args, match_against=["comments"], **kwargs
                )
            },
        )

    def emptyline(self):
        # Don't repeat last command
        return ""

    def get_names(self):
        # Prevent from having «EOF» in commands list
        names = [
            name
            for name in Cmd.get_names(self)
            if name not in ("do_eof", "do_eos", "do__relative_load")
        ]
        return names

    def run(self):
        """Run the command shell"""

        intro = "Welcome to the EeSSH shell. Type help or ? to list commands.\n"

        loop = True
        while loop:
            try:
                self.cmdloop(intro=intro)

            except KeyboardInterrupt:
                intro = "\n"

            else:
                loop = False

        return self.lastcmd
