# EeSSH -- SSH client for Eemyle
# By: Cyril Lacoux <clacoux@easter-eggs.com>
#
# Copyright (C) 2003-2019 Easter-eggs
# https://gitlab.easter-eggs.com/ee/eessh
#
# This file is part of EeSSH.
#
# EeSSH is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# EeSSH is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helpers collection"""

import datetime
import errno
import logging
import os
import pydoc
import re
import shlex
import shutil
import socket
import stat
import threading
from pathlib import Path

import prettytable

from eessh.api import api
from eessh.config import config

log = logging.getLogger(__name__)


KB = 1024
MB = KB * KB
GB = MB * KB
TB = GB * KB
PB = TB * KB


class MasterFifo(threading.Thread):
    """Master FIFO object (used for EeSSH broadcast feature)"""

    type = "master"

    def __init__(self, filepath):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.bind(filepath)

        os.chmod(filepath, 0o600)

        self.__clients = []
        self.__filepath = filepath
        self.__loop = False

        # Calling inherited
        super().__init__()

        # Non-blocking mode
        self.daemon = True

    def close(self):
        """Close FIFO"""
        self.__loop = False

        self.sock.close()

        os.remove(self.__filepath)

    stop = close

    def run(self):
        """Start FIFO handler"""
        self.__loop = True

        self.sock.listen(1)
        while self.__loop:
            client, info = self.sock.accept()  # pylint: disable=unused-variable
            self.__clients.append(client)

    def send(self, data):
        """Send data"""
        for client in self.__clients[:]:
            try:
                client.send(data)
            except OSError as err:
                if err.errno == errno.EPIPE:
                    log.info("[BROADCAST] Client closed connection")
                    self.__clients.remove(client)
                    continue

                log.exception("[BROADCAST] failed to send data to client")


class SlaveFifo(socket.socket):
    """Slave FIFO object (used for EeSSH broadcast feature)"""

    type = "slave"

    def __init__(self, filepath):
        # Calling inherited
        super().__init__(socket.AF_UNIX, socket.SOCK_STREAM)

        self.__filepath = filepath

    def read(self, length):
        """Read data"""
        return self.recv(length)

    def start(self):
        """Start FIFO"""
        self.connect(self.__filepath)

    def stop(self):
        """Stop FIFO"""
        return self.close()

    def write(self, data):
        """Write data"""
        return self.send(data)


def attr_data(attr):
    """Compute and format path attributes from the result of SFTP stat() method"""

    def rwx(node, suid, sticky=False):
        if suid:
            suid = 2
        out = "-r"[node >> 2] + "-w"[(node >> 1) & 1]
        if sticky:
            out += "-xTt"[suid + (node & 1)]
        else:
            out += "-xSs"[suid + (node & 1)]
        return out

    data = {}
    if attr.st_mode is not None:
        kind = stat.S_IFMT(attr.st_mode)
        if kind == stat.S_IFIFO:
            data["type"] = "p"
        elif kind == stat.S_IFCHR:
            data["type"] = "c"
        elif kind == stat.S_IFDIR:
            data["type"] = "d"
        elif kind == stat.S_IFBLK:
            data["type"] = "b"
        elif kind == stat.S_IFREG:
            data["type"] = "-"
        elif kind == stat.S_IFLNK:
            data["type"] = "l"
        elif kind == stat.S_IFSOCK:
            data["type"] = "s"
        else:
            data["type"] = "?"

        data["mode"] = attr.st_mode

        data["owner_permissions"] = rwx((attr.st_mode & 0o700) >> 6, attr.st_mode & stat.S_ISUID)
        data["group_permissions"] = rwx((attr.st_mode & 0o70) >> 3, attr.st_mode & stat.S_ISGID)
        data["other_permissions"] = rwx(attr.st_mode & 7, attr.st_mode & stat.S_ISVTX, True)

    else:
        data["type"] = "?"
        data["permissions"] = None
        data["owner_permissions"] = "---"
        data["group_permissions"] = "---"
        data["other_permissions"] = "---"

    # compute display date
    if (attr.st_mtime is None) or (attr.st_mtime == 0xFFFFFFFF):
        # shouldn't really happen
        data["mtime"] = None
    else:
        data["mtime"] = datetime.datetime.fromtimestamp(attr.st_mtime)

    data["filename"] = getattr(attr, "filename", "?")

    # not all servers support uid/gid
    data["uid"] = attr.st_uid or 0
    data["gid"] = attr.st_gid or 0

    data["size"] = attr.st_size or 0

    return data


def attr_repr(data):
    """
    Slightly modified version of SFTPAttributes.__str__() with unicode support and human readable size
    """

    data = data.copy()

    data["size_hr"] = format_size(data["size"])
    data["mtime_hr"] = format_mtime(data["mtime"])
    data["link_info"] = f' -> {data["link_attr"]["path"]}' if data["type"] == "l" else ""

    # pylint: disable=consider-using-f-string
    return "{type}{owner_permissions}{group_permissions}{other_permissions} 1 {uid:<8d} {gid:<8d} {size_hr:>12s} {mtime_hr:>15s} {filename}{link_info}".format(
        **data
    )


def attr_sort(attr1, attr2):
    """Sort function for SFTP attributes objects"""

    # Directory first
    if attr1["type"] == "d":
        if not attr2["type"] == "d":
            return -1

    elif attr2["type"] == "d":
        return 1

    filename1 = attr1["filename"].lower()
    filename2 = attr2["filename"].lower()

    return (filename1 > filename2) - (filename1 < filename2)


def format_mtime(date):
    """Format modification time"""
    today = datetime.datetime.today()
    delta = today - date

    if abs(delta.days) > 182:
        # 6 months
        format_ = "%-2d %5b %-5Y"
    else:
        format_ = "%-2d %5b %H:%M"

    return date.strftime(format_)


def get_host_port(spec):
    """Split a hostname:port string"""
    if ":" in spec:
        host, port = spec.split(":")

    else:
        host = "127.0.0.1"
        port = spec

    port = int(port)
    return (host, port)


def format_size(size):
    """Format size specified as bytes"""
    if not size:
        return "0"

    if size >= PB:
        return f"{size / PB:7.2f}P"

    if size >= TB:
        return f"{size / TB:7.2f}T"

    if size >= GB:
        return f"{size / GB:7.2f}G"

    if size >= MB:
        return f"{size / MB:7.2f}M"

    if size >= KB:
        return f"{size / KB:7.2f}K"

    return f"{size}B"


def format_time(duration):
    """Format a duration specified in seconds"""
    if duration > 86400:
        # > 1d
        return "STALLED"

    parts = []

    # hours
    if duration > 3600:
        parts.append(f"{duration // 3600:02d}")
        duration = duration % 3600

    # minutes
    parts.append(f"{duration // 60:02d}")
    duration = duration % 60

    # seconds
    parts.append(f"{duration:02d}")

    return ":".join(parts)


def get_terminal_size(default_witdh=None, default_height=None):
    """Retrieve terminal size"""
    return shutil.get_terminal_size(
        (default_witdh if default_witdh else 80, default_height if default_height else 80)
    )


def wrap_text(text, width=None, splited=False, global_prefix=None):
    """Wrap text for output"""
    width = width if width else get_terminal_size().columns
    global_prefix = global_prefix if global_prefix else ""
    result = []
    for srcline in text.splitlines():
        # Replace tabs by 4 spaces
        srcline = srcline.replace("\t", "    ")

        # Handle case of really long line
        if len(srcline) <= width:
            result.append(global_prefix + srcline)
            continue

        # Detect prefix
        matches = re.match("^( *)(.*)$", srcline)
        prefix = matches.group(1)
        line_limit = width - len(prefix) - len(global_prefix)
        line = ""
        for word in matches.group(2).split(" "):
            # Handle case of really long word
            if len(word) > line_limit:
                if line:
                    result.append(global_prefix + prefix + line)
                line = ""
                result.append(global_prefix + prefix + word)
                continue

            # Switch to new line if line limit exceeded with this word
            if (len(line) + len(word) + 1) > line_limit:
                result.append(global_prefix + prefix + line)
                line = ""

            # Finally, append word to current line
            line = word if not line else line + " " + word
        if line:
            result.append(global_prefix + prefix + line)
    return result if splited else "\n".join(result)


def pretty_output_line(info, value, width=None, splited=False, line_return=True):
    """Format one line for output"""
    width = width if width else get_terminal_size().columns
    value = "not set" if value is None else str(value)
    result = [f"{info}: {value}"]
    if len(result[0]) > width or "\n" in value:
        result = [info + ":\n"]
        result.extend(wrap_text(value, width=width, global_prefix="  ", splited=True))
    if splited:
        return result
    result = "\n".join(result)
    return result if not line_return else result + "\n"


def pretty_output_lines(lines, width=None, splited=False, line_return=True):
    """Format lines for output"""
    width = width if width else get_terminal_size().columns
    result = []
    for line in lines:
        if isinstance(line, str):
            result.append(line)
        else:
            result.extend(pretty_output_line(*line, width=width, splited=True))
    if splited:
        return result
    result = "\n".join(result)
    return result if not line_return else result + "\n"


def auto_paged(text):
    """Automatically enable paging if content is larger than the terminal height"""
    if len(text.split("\n")) > get_terminal_size().lines:
        pydoc.pager(text)
    print(text)


def pretty_format_server_list(servers):
    """Format list of server in an ASCII table"""
    table = prettytable.PrettyTable()
    if hasattr(prettytable, "DOUBLE_BORDER"):
        table.set_style(prettytable.DOUBLE_BORDER)  # pylint: disable=no-member
    table.field_names = ["Hostname", "OS", "Description"]
    table.align["Hostname"] = "l"  # pylint: disable=unsupported-assignment-operation
    table.align["Description"] = "l"  # pylint: disable=unsupported-assignment-operation
    for server in servers:
        table.add_row(
            [server["hostname"], server["os_name"], wrap_text(server["description"], width=40)]
        )
    return table.get_string(sortby="Hostname")


def pretty_format_server_tickets(tickets):
    """Format server tickets in an ASCII table"""
    table = prettytable.PrettyTable()
    if hasattr(prettytable, "DOUBLE_BORDER"):
        table.set_style(prettytable.DOUBLE_BORDER)  # pylint: disable=no-member
    table.field_names = ["ID", "Subject", "Creation date", "Last update", "Owner", "Status"]
    table.align["Subject"] = "l"  # pylint: disable=unsupported-assignment-operation
    for ticket in tickets:
        table.add_row(
            [
                ticket["number"],
                ticket["subject"],
                ticket["date"],
                ticket["last_updated"],
                ticket["owner"],
                ticket["status"],
            ]
        )
    return table.get_string(sortby="Last update", reversesort=True)


def pretty_format_server_data(data):
    """Format server extended data"""
    return pretty_output_lines(
        [
            ("Hostname", f'{data["hostname"]} (#{data["id"]})'),
            ("Customer UID", data["customer_uid"]),
            ("Description", data["description"]),
            ("Security update", "yes" if data["security"] else "no"),
            "Interfaces: %s"
            % (
                "no interface"
                if not data["interfaces"]
                else "\n - " + "\n - ".join([interface["name"] for interface in data["interfaces"]])
            ),
            (
                "Physical host",
                "not set" if not data["physical_host"]
                # pylint: disable=consider-using-f-string
                else "{hostname} ({customer_uid})".format(**data["physical_host"]),
            ),
            (
                "Connection on",
                "not set"
                if not data["connection_interface"]
                else data["connection_interface"]["name"],
            ),
            "Connection hops: %s"
            % (
                "no hop"
                if not data["hops"]
                else "\n - "
                + "\n - ".join(
                    # pylint: disable=consider-using-f-string
                    ["{username}@{hostname} ({host}:{port})".format(**hop) for hop in data["hops"]]
                )
            ),
            (
                "Users",
                "no user"
                if not data["users"]
                else ", ".join([user["username"] for user in data["users"]]),
            ),
            (
                "IPMI",
                "not configured" if not data["ipmi"]["enabled"]
                # pylint: disable=consider-using-f-string
                else "{username}@{host}:{port}".format(**data["ipmi"]),
            ),
            ("Comments", "no comment" if not data["comments"] else data["comments"]),
        ]
    )


def pretty_format_server_passwords(data):
    """Format server usernames & passords in an ASCII table"""
    table = prettytable.PrettyTable()
    if hasattr(prettytable, "DOUBLE_BORDER"):
        table.set_style(prettytable.DOUBLE_BORDER)  # pylint: disable=no-member
    table.field_names = ["Username", "Password"]
    table.align = "l"
    for user in data["users"]:
        table.add_row([user["username"], user["password"] if user["password"] else ""])
    return table.get_string(sortby="Username")


def pretty_format_pfwd(pfwds, reverse=False):
    """Format ports fordarding collection in an ASCII table"""
    table = prettytable.PrettyTable()
    if hasattr(prettytable, "DOUBLE_BORDER"):
        table.set_style(prettytable.DOUBLE_BORDER)  # pylint: disable=no-member
    if reverse:
        table.field_names = ["Index", "Remote host", "Remote port", "Local host", "Local port"]
    else:
        table.field_names = ["Index", "Local host", "Local port", "Remote host", "Remote port"]
    table.align = "l"
    for index, h1, p1, h2, p2 in pfwds:  # pylint: disable=invalid-name
        table.add_row([index, h1, p1, h2, p2])
    return table.get_string(sortby="Index")


def resolv_local_paths(raw_path):
    """Resolve local paths pattern to list of existing pathlib.Path objects"""
    if raw_path.startswith("~"):
        homeref = raw_path.split(os.path.sep)[0]
        raw_path = os.path.join(Path(homeref).expanduser(), raw_path[len(homeref)])
    if "*" not in raw_path:
        return [Path(raw_path)]
    if raw_path.startswith(os.path.sep):
        return list(Path(os.path.sep).glob(raw_path[len(os.path.sep) :]))
    return list(Path(".").glob(raw_path))


def generate_server_ansible_inventory_entry(
    customer_uid, hostname, data=None, no_firsthop=False, refresh=False
):
    """Generate ansible inventory entry for the specified server"""
    data = api.get_data(customer_uid, hostname, refresh=refresh) if data is None else data
    if not data:
        log.error("Failed to get data for server %s (%s)", hostname, customer_uid)
        return False
    if not data["hops"]:
        log.error("No hop retrieve for %s (%s)", hostname, customer_uid)
        return False
    host = data["hops"][-1]

    entry = {
        "ansible_host": host["host"],
        "ansible_port": host["port"],
        "ansible_user": host["username"],
        "ansible_password": host["password"],
        "ansible_become_method": host["su_method"],
        "ansible_become_user": host["su_username"],
        "ansible_become_password": host["su_password"],
    }
    # Remove undefined/empty keys
    entry = {key: value for key, value in entry.items() if value}

    hops = data["hops"][0:-1]
    if config.get("firsthop", "enabled") and not no_firsthop:
        hops.insert(
            0,
            {
                "host": config.get("firsthop", "hostname"),
                "port": config.get("firsthop", "port"),
                "username": config.get("firsthop", "username"),
            },
        )

    if hops:
        # pylint: disable=consider-using-f-string
        entry["ansible_ssh_common_args"] = '-J "%s"' % ",".join(
            # pylint: disable=consider-using-f-string
            ["{username}@{host}:{port}".format(**hop) for hop in hops]
        )

    # pylint: disable=consider-using-f-string
    return "{}-{} {}".format(
        customer_uid, hostname, " ".join([f"{key}='{value}'" for key, value in entry.items()])
    )


def generate_server_raw_ssh_command(
    customer_uid, hostname, data=None, no_firsthop=False, refresh=False
):
    """Generate raw SSH command to connect on the specified server"""
    data = api.get_data(customer_uid, hostname, refresh=refresh) if data is None else data
    if not data:
        log.error("Failed to get data for server %s (%s)", hostname, customer_uid)
        return False
    if not data["hops"]:
        log.error("No hop retrieve for %s (%s)", hostname, customer_uid)
        return False
    host = data["hops"][-1]

    command = ["ssh"]
    command.append(f"{host['username']}@{host['host']}" if host["username"] else host["host"])
    if host["port"]:
        command.extend(["-p", str(host["port"])])

    hops = data["hops"][0:-1]
    if config.get("firsthop", "enabled") and not no_firsthop:
        hops.insert(
            0,
            {
                "host": config.get("firsthop", "hostname"),
                "port": config.get("firsthop", "port"),
                "username": config.get("firsthop", "username"),
            },
        )

    if hops:
        command.extend(
            [
                "-J",
                ",".join([f"{hop['username']}@{hop['host']}:{hop['port']}" for hop in hops]),
            ]
        )

    return shlex.join(command)
